from pedal.core.commands import gently
from pedal.toolkit.utilities import find_function_calls, is_top_level


def ensure_prints(count):
    """

    Args:
        count:

    Returns:

    """
    prints = find_function_calls('print')
    if not prints:
        gently("You are not using the print function!", label="no_print", title="Missing Print")
        return False
    elif len(prints) > count:
        gently("You are printing too many times!", label="multiple_print", title="Too Many Prints")
        return False
    elif len(prints) < count:
        gently("You are not printing enough things!", label="too_few_print", title="Too Few Prints")
        return False
    else:
        for a_print in prints:
            if not is_top_level(a_print):
                gently("You have a print function that is not at the top level. That is incorrect for this problem!",
                       label="not_top_level_print", title="Non-Top Level Print")
                return False
    return prints



import re
from pedal.source import get_program
from pedal.sandbox.compatibility import get_output
from pedal.core.commands import gently, explain

# TODO: Move this into the CS1014 package!


# Feedback for author's name
def check_author_name_on_header():
    """

    """
    code = get_program()
    m_author = re.search('Author: \\w+', code)
    if not m_author:
        gently("You need to add your name to the author field at the top of the file.", label="name_missing",
                 title="Missing Name")


def get_plots(output):
    """

    Args:
        output:

    Returns:

    """
    # The p[0] is the first plot in a graph/show
    return [p[0] for p in output if isinstance(p[0], dict)]


def find_plot_of_type(plot_list, plot_type):
    """

    Args:
        plot_list:
        plot_type:

    Returns:

    """
    return [p['data'] for p in plot_list if p['type'] == plot_type]


# Feedback for copying output of the program in the documentation
def check_output_on_header(expected_output):
    """

    Args:
        expected_output:
    """
    code = get_program()
    expected_output = str(expected_output)
    between_stars = code.split("*****")[2].strip()
    between_stars = "\\n".join([x.strip() for x in between_stars.split("\\n")])
    if 'REPLACE THIS TEXT WITH THE OUTPUT OF THIS PROGRAM' in between_stars:
        gently("In your code, you need to 'REPLACE THIS TEXT WITH THE OUTPUT OF THIS PROGRAM'",
               label="wrong_output_blank",
               title="Blank Output")
    elif expected_output not in between_stars:
        gently("The output you copied between the *****, seems to be incorrect. "
               "You may have copied it into the wrong location, or it is incomplete.",
               label="wrong_output_fill", title="Missing Output")


def check_problem_submission(prob_id):
    """

    Args:
        prob_id:

    Returns:

    """
    if prob_id not in get_program():
        explain("Make sure that you are turning in {}</br>".format(prob_id),
                label="wrong_problem", title="Wrong Problem")
        return True


def check_print_output(multiple_lines):
    """

    Args:
        multiple_lines:

    Returns:

    """
    for line in multiple_lines:
        if line not in get_output():
            gently("You are not doing the correct calculation</br>", label="catch_all", title="Wrong Output")
            return True


def find_in_code(regex):
    """

    Args:
        regex:

    Returns:

    """
    code = get_program()
    return re.search(regex, code)



from pedal.core.report import MAIN_REPORT
from pedal.core.commands import set_success
from pedal.core.feedback import Feedback
from pedal.resolvers.core import make_resolver

DEFAULT_NO_FEEDBACK_MESSAGE = "No errors reported."
DEFAULT_NO_FEEDBACK_TITLE = "No Errors"


DEFAULT_CATEGORY_PRIORITY = [
    "highest",
    Feedback.CATEGORIES.SYNTAX,
    Feedback.CATEGORIES.MISTAKES,
    Feedback.CATEGORIES.INSTRUCTOR,
    Feedback.CATEGORIES.ALGORITHMIC,
    Feedback.CATEGORIES.RUNTIME,
    Feedback.CATEGORIES.STUDENT,
    Feedback.CATEGORIES.POSITIVE,
    Feedback.CATEGORIES.INSTRUCTIONS,
    Feedback.CATEGORIES.UNKNOWN,
    "lowest"
]


class FinalFeedback:
    """
    Internal class used for organizing the feedback into one place. Dumpable into a simple dictionary.
    """
    def __init__(self, success=None, score=None, category=None, label=None, title=None,
                 message=None, data=None, hide_correctness=None):
        self.success = success
        self.score = score
        self.category = category
        self.label = label
        self.title = title
        self.message = message
        self.data = data
        self.hide_correctness = hide_correctness

    def __str__(self) -> str:
        return "FinalFeedback({label!r}, {title!r}, {message!r})".format(label=self.label,
                                                                         title=self.title,
                                                                         message=self.message[:50])

    def for_console(self) -> str:
        return "{label}\n{score}\n{title}\n{message}".format(label=self.label,
                                                             score=self.score,
                                                             title=self.title,
                                                             message=self.message)

    def to_json(self) -> dict:
        """

        Returns:

        """
        return {
            'success': self.success,
            'score': self.score,
            'category': self.category,
            'label': self.label,
            'title': self.title,
            'message': self.message,
            'data': self.data,
            'hide_correctness': self.hide_correctness
        }


def by_priority(feedback):
    """
    Converts a feedback into a numeric representation for sorting.

    Args:
        feedback (Feedback): The feedback object to convert
    Returns:
        float: A decimal number representing the feedback's relative priority.
    """
    category = Feedback.CATEGORIES.UNKNOWN
    if feedback.category is not None:
        category = feedback.category.lower()
    priority = 'medium'
    if feedback.priority is not None:
        priority = feedback.priority.lower()
        priority = Feedback.CATEGORIES.ALIASES.get(priority, priority)
    if category in DEFAULT_CATEGORY_PRIORITY:
        value = DEFAULT_CATEGORY_PRIORITY.index(category)
    else:
        value = len(DEFAULT_CATEGORY_PRIORITY)
    if priority in DEFAULT_CATEGORY_PRIORITY:
        value = DEFAULT_CATEGORY_PRIORITY.index(priority)
    offset = priority_offset(priority)
    return value + offset


def priority_offset(priority):
    """

    Args:
        priority:

    Returns:

    """
    if priority == 'low':
        return .7
    elif priority == 'medium':
        return .5
    elif priority == 'high':
        return .3
    else:
        return .1


def parse_feedback(feedback):
    """

    Args:
        feedback:

    Returns:

    """
    message = feedback.message or feedback.text
    title = feedback.title or feedback.label
    return feedback.correct, feedback.score, message, title, feedback.fields


@make_resolver
def resolve(report=MAIN_REPORT, priority_key=by_priority):
    """
    Args:
        priority_key: The key function to sort feedbacks by
        report (Report): The report object to resolve down. Defaults to the
                         global MAIN_REPORT

    Returns
        str: A string of HTML feedback to be delivered
    """
    # Prepare feedbacks
    feedbacks = report.feedback
    feedbacks.sort(key=priority_key)
    suppressions = report.suppressions
    # Process
    final = FinalFeedback(success=False, score=0,
                          title=None, message=None,
                          category=Feedback.CATEGORIES.COMPLETE, label='set_success_no_errors',
                          data=[], hide_correctness=False)
    for feedback in feedbacks:
        category = feedback.category.lower()
        if category in suppressions:
            if True in suppressions[category]:
                continue
            elif feedback.label.lower() in suppressions[category]:
                continue
        if feedback.label in report.suppressed_labels:
            continue
        success, partial, message, title, data = parse_feedback(feedback)
        final.score += partial if partial is not None else 0
        if feedback.muted:
            continue
        final.success = success or final.success
        if message is not None and final.message is None and feedback.priority != 'positive':
            final.message = message
            final.title = title
            final.category = feedback.category
            final.label = feedback.label
            final.data = data
    if final.message is None:
        final.title = DEFAULT_NO_FEEDBACK_TITLE
        final.message = DEFAULT_NO_FEEDBACK_MESSAGE
    final.hide_correctness = suppressions.get('success', False)
    if (not final.hide_correctness and final.success and
            final.label == 'set_success_no_errors' and final.category == Feedback.CATEGORIES.COMPLETE):
        # TODO: Promote to be its own atomic feedback function
        final.title = set_success.title
        final.message = set_success.text_template()
    report.result = final
    return final



from pedal.cait.cait_api import parse_program
from pedal.cait.cait_node import CaitNode
from pedal.core.commands import gently, explain
from pedal.core.feedback import AtomicFeedbackFunction, CompositeFeedbackFunction
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.utilities.operators import COMPARE_OP_NAMES, BIN_OP_NAMES, BOOL_OP_NAMES, UNARY_OP_NAMES


def is_top_level(ast_node: CaitNode, report=MAIN_REPORT) -> bool:
    """
    Determines if the `ast_node` is at the top-level of the program.
    Correctly handles expression statements (so a print call on its own will be
    considered a statement, even though its technically an expression).

    Args:
        ast_node (CaitNode): The CaitNode to check

    Returns:
        bool: Whether the node is from the top level
    """
    ast = parse_program(report=MAIN_REPORT)
    for element in ast.body:
        if element.ast_name == "Expr":
            if element.value == ast_node:
                return True
        elif element == ast_node:
            return True
    return False


@AtomicFeedbackFunction(
    message_template="You have defined a function inside of another block. For instance, you may "
    "have placed it inside another function definition, or inside of a loop. Do not nest your function "
    "definition!",
    title="Don't Nest Functions",
    justification="Found a FunctionDef that was not at the top-level.",
)
def no_nested_function_definitions(muted=False, report=MAIN_REPORT) -> bool:
    """
    Returns `True` if there are any functions defined inside of other functions.
    Also attaches feedback, although that can be muted.

    Returns:
        bool: Returns whether there is a nested definition
    """
    ast = parse_program(report=MAIN_REPORT)
    defs = ast.find_all("FunctionDef")
    for a_def in defs:
        if not is_top_level(a_def):
            gently(
                no_nested_function_definitions.text_template,
                label="no_nested_function_definitions",
                muted=muted,
                report=MAIN_REPORT,
            )
            return False
    return True


def function_prints(function_name: str = None, report=MAIN_REPORT) -> bool:
    """
    Determines if there is a print statement inside of any functions. If `function_name` is given,
    then only that function will be checked.

    Returns:
        bool: Whether there is a print inside the function.
    """
    ast = parse_program(report=report)
    defs = ast.find_all("FunctionDef")
    for a_def in defs:
        if function_name is not None and a_def.name != function_name:
            continue
        all_calls = a_def.find_all("Call")
        for a_call in all_calls:
            if a_call.func.ast_name == "Name":
                if a_call.func.id == "print":
                    return True
    return False


def find_function_calls(name: str, root=None, report=MAIN_REPORT):
    """
    Returns a list of CaitNodes representing all of the function calls that
    were found. This includes both methods and regular functions.

    Args:
        name (str): The name of the function to search.
        root: A subtree of a parse tree to revert to.
        report (Report): The name of the Report to refer to.

    Returns:
        List[CaitNode]: Relevant call nodes.
    """
    if root is None:
        root = parse_program(report=report)
    all_calls = root.find_all("Call")
    calls = []
    for a_call in all_calls:
        if a_call.func.ast_name == "Attribute":
            if a_call.func.attr == name:
                calls.append(a_call)
        elif a_call.func.ast_name == "Name":
            if a_call.func.id == name:
                calls.append(a_call)
    return calls


def function_is_called(name: str, report=MAIN_REPORT) -> int:
    """
    Returns the number of times that the given function name is called.
    """
    return len(find_function_calls(name, report=report))


def only_printing_variables(report=MAIN_REPORT) -> bool:
    """
    Determines whether the user is only printing variables, as opposed to
    literal values.

    """
    ast = parse_program(report=report)
    all_calls = ast.find_all("Call")
    for a_call in all_calls:
        if a_call.func.ast_name == "Name" and a_call.func.id == "print":
            for arg in a_call.args:
                if arg.ast_name != "Name":
                    return False
                elif arg.id in ("True", "False", "None"):
                    return False
    return True


def find_prior_initializations(node, report=MAIN_REPORT) -> [CaitNode]:
    """

    Given a Name node, returns a list of all the assignment
    statements that incorporate that Name node prior to that line. Returns
    None if no Name is given.

    """
    if node.ast_name != "Name":
        return None
    ast = parse_program(report=report)
    assignments = ast.find_all("Assign")
    cur_line_no = node.lineno
    all_assignments = []
    for assignment in assignments:
        if assignment.has(node):
            if assignment.lineno < cur_line_no:
                all_assignments.append(assignment)
    return all_assignments


# TODO: Fill in other returning methods, add in returning funtions
RETURNING_METHODS = (
    "replace",
    "strip",
    "lstrip",
    "rstrip",
    "split",
    "upper",
    "lower",
    "title",
)


@AtomicFeedbackFunction(
    title="Did Not Use Function's Return Value",
    message_template="It looks like you called the {kind} `{name}` on {location.line}, but failed to store the result in a variable or use it in an expression. You should remember to use the result!",
)
def prevent_unused_result(
    muted=False, report=MAIN_REPORT, returning_methods=RETURNING_METHODS
) -> [CaitNode]:
    """

    Returns a list of any function calls where the function being called
    typically has a return value that should be assigned or used in an
    expression, but was instead thrown away.

    """
    ast = parse_program(report=report)
    exprs = ast.find_all("Expr")
    returns = []
    for expr in exprs:
        if expr.value.ast_name == "Call":
            kind = "function"
            continue
        else:
            kind = "method"
        a_call = expr.value
        if a_call.func.ast_name != "Attribute":
            continue
        method_name = a_call.func.attr
        if method_name == "append":
            pass
        elif method_name in prevent_unused_result.methods:
            fields = {
                "location": Location.from_ast(expr),
                "kind": kind,
                "name": a_call.func.attr,
            }
            gently(
                prevent_unused_result.message_template.format(**fields),
                label=prevent_unused_result.label,
                title=prevent_unused_result.title,
                fields=fields,
                muted=muted, report=report
            )
            returns.append(a_call)
    return returns


@AtomicFeedbackFunction(
    title="Builtin Usage",
    message_template="You cannot use the builtin function `{name}`.",
)
def prevent_builtin_usage(function_names: [str], muted: bool = False, report=MAIN_REPORT) -> str:
    """

    Determines the name of the first function in `function_names` that is
    called, or returns `None`; also attaches feedback.

    """
    # Prevent direction calls
    ast = parse_program(report=report)
    all_calls = ast.find_all("Call")
    for a_call in all_calls:
        if a_call.func.ast_name == "Name":
            if a_call.func.id in function_names:
                fields = {"location": Location.from_ast(a_call), "name": a_call.func.id}
                explain(
                    prevent_builtin_usage.message_template.format(**fields),
                    fields=fields,
                    muted=muted,
                    label=prevent_builtin_usage.label,
                    title=prevent_builtin_usage.title,
                )
                return a_call.func.id
    return None


def find_negatives(root=None, report=MAIN_REPORT):
    """

    Returns all the occurrences of the given literal negative number in the source
    code. Can optionally filter at the given subtree.

    """
    if root is None:
        root = parse_program(report=report)
    return [
        -op.operand.n
        for op in root.find_all("UnaryOp")
        if op.op.ast_name == "USub" and op.operand.ast_name == "Num"
    ]


@AtomicFeedbackFunction(title="Do Not Use Literal Value")
def prevent_literal(*literals, muted=False, report=MAIN_REPORT):
    """
    Confirms that the literal is not in the code, returning False if it is not.
    
    Args:
        *literals (Any...): A series of literal values to look for.
    Returns:
        AstNode or False: If the literal is found in the code, then it is returned.
    """
    message = "Do not use the literal value <code>{}</code> in your code."
    ast = parse_program(report=report)
    str_values = [s.s for s in ast.find_all("Str")]
    num_values = [n.n for n in ast.find_all("Num")]
    negative_values = find_negatives(ast, report=report)
    name_values = [name.id for name in ast.find_all("Name")] + [
        name.value for name in ast.find_all("NameConstant")
    ]
    for literal in literals:
        if isinstance(literal, (int, float)):
            if literal in num_values or literal in negative_values:
                if not muted:
                    explain(message.format(repr(literal)), label=prevent_literal.label, title=prevent_literal.title, muted=muted, report=report)
                return literal
        elif isinstance(literal, str):
            if literal in str_values:
                if not muted:
                    explain(message.format(repr(literal)), label=prevent_literal.label, title=prevent_literal.title, muted=muted, report=report)
                return literal
        elif literal in (True, False, None):
            if str(literal) in name_values:
                if not muted:
                    explain(message.format(repr(literal)), label=prevent_literal, title=prevent_literal.title, muted=muted, report=report)
                return literal
    return False


@AtomicFeedbackFunction(title="Missing Literal")
def ensure_literal(*literals, muted=False, report=MAIN_REPORT):
    """
    Confirms that the literal IS in the code, returning False if it is not.
    
    Args:
        *literals (Any...): A series of literal values to look for.
    Returns:
        AstNode or False: If the literal is found in the code, then it is returned.
    """
    message = "You need the literal value <code>{}</code> in your code."
    ast = parse_program(report=report)
    str_values = [s.s for s in ast.find_all("Str")]
    num_values = [n.n for n in ast.find_all("Num")]
    negative_values = find_negatives(ast, report=report)
    name_values = [str(name.id) for name in ast.find_all("Name")] + [
        str(name.value) for name in ast.find_all("NameConstant")
    ]
    for literal in literals:
        if literal in (True, False, None):
            if str(literal) not in name_values:
                if not muted:
                    explain(message.format(repr(literal)), label=ensure_literal.label, title=ensure_literal.title, muted=muted, report=report)
                return True
        elif isinstance(literal, (int, float)):
            if literal not in num_values and literal not in negative_values:
                if not muted:
                    explain(message.format(repr(literal)), label=ensure_literal.label, title=ensure_literal.title, muted=muted, report=report)
                return literal
        elif isinstance(literal, str):
            if literal not in str_values:
                if not muted:
                    explain(message.format(repr(literal)), label=ensure_literal.label, title=ensure_literal.title, muted=muted, report=report)
                return literal
    return False


def prevent_advanced_iteration(muted=False, report=MAIN_REPORT):
    """

    """
    message = "You should not use a <code>while</code> loop to solve this problem."
    label = "while_usage"
    title = "Usage of <code>while</code>"
    ast = parse_program()
    if ast.find_all("While"):
        explain(message, label=label, title=title, muted=muted, report=report)
    return prevent_builtin_usage(
        [
            "sum",
            "map",
            "filter",
            "reduce",
            "len",
            "max",
            "min",
            "max",
            "sorted",
            "all",
            "any",
            "getattr",
            "setattr",
            "eval",
            "exec",
            "iter",
        ], muted=muted, report=report
    )


@AtomicFeedbackFunction(title="Missing Operator")
def ensure_operation(op_name, root=None, muted=False, report=MAIN_REPORT):
    """

    Determines if the given operator `op_name` is used anywhere, returning the
    node of it if it is. Otherwise, returns `None`. You can specify the operator
    as a string like `"+"` or `"<<"`. Supports all comparison, boolean, binary, and unary operators.

    """
    message = "You are not using the `{name}` operator.".format(name=op_name)
    if root is None:
        root = parse_program()
    result = find_operation(op_name, root)
    if not result:
        gently(message, label=ensure_operation.label,
               title=ensure_operation.title, muted=muted, report=report)
    return result


@AtomicFeedbackFunction(title="May Not Use Operator",
                        message_template="You may not use the `{name}` operator.")
def prevent_operation(op_name, root=None, muted=False, report=MAIN_REPORT):
    """

    Determines if the given operator `op_name` is not used anywhere, returning the
    node of it if it is. Otherwise, returns `None`. You can specify the operator
    as a string like `"+"` or `"<<"`. Supports all comparison, boolean, binary, and unary operators.

    """
    root = root or parse_program()
    result = find_operation(op_name, root)
    if result:
        fields = {'name': op_name, 'location': Location.from_ast(result)}
        gently(prevent_operation.message_template.format(**fields),
               label=prevent_operation.label, title=prevent_operation.title,
               fields=fields,
               muted=muted, report=report)
    return result


def find_operation(op_name, root=None, report=MAIN_REPORT):
    """

    Returns the first occurrence of the operator `op_name` in the source code.
    Otherwise returns `False`. You can specify the operator
    as a string like `"+"` or `"<<"`. Supports all comparison, boolean, binary, and unary operators.

    """
    root = root or parse_program()
    if op_name in COMPARE_OP_NAMES:
        compares = root.find_all("Compare")
        for compare in compares:
            for op in compare.ops:
                if op.ast_name == COMPARE_OP_NAMES[op_name]:
                    return compare
    elif op_name in BOOL_OP_NAMES:
        boolops = root.find_all("BoolOp")
        for boolop in boolops:
            if boolop.op_name == BOOL_OP_NAMES[op_name]:
                return boolop
    elif op_name in BIN_OP_NAMES:
        binops = root.find_all("BinOp")
        for binop in binops:
            if binop.op_name == BIN_OP_NAMES[op_name]:
                return binop
    elif op_name in UNARY_OP_NAMES:
        unaryops = root.find_all("UnaryOp")
        for unaryop in unaryops:
            if unaryop.op_name == UNARY_OP_NAMES[op_name]:
                return unaryop
    return False


@CompositeFeedbackFunction()
def ensure_assignment(variable_name, type=None, value=None, root=None, muted=False, report=MAIN_REPORT):
    """

    Consumes a variable name
    TODO: Implement the value parameter

    :param variable_name: The variable name the student is expected to define.
    :type variable_name: str
    :param type: The string type of the node on the right side of the
                 assignment. Check GreenTreeSnakes (e.g., "Num", or "Str").
    :type type: str
    :return: False or str

    Args:
        root:
        value:
    """
    if root is None:
        root = parse_program()
    assignments = root.find_all("Assign")
    potentials = []
    for assign in assignments:
        if assign.targets[0].ast_name != "Name":
            continue
        if assign.targets[0].id == variable_name:
            potentials.append(assign)
            if type is None:
                return assign
            elif (
                type == "Bool"
                and assign.value.ast_name == "Name"
                and assign.value.id in ("True", "False")
            ):
                return assign
            elif (
                type == "Bool"
                and assign.value.ast_name == "NameConstant"
                and assign.value.value in (True, False)
            ):
                return assign
            elif assign.value.ast_name == type:
                return assign
    if potentials and potentials[0].value.ast_name not in (
        "Str",
        "Bool",
        "Num",
        "List",
        "Tuple",
    ):
        explain(
            (
                "You needed to assign a literal value to {variable}, but you "
                "created an expression instead."
            ).format(variable=variable_name),
            label="exp_vs_lit",
            title="Expression Instead of Literal",
            report=report, muted=muted
        )
    elif type is None:
        explain(
            (
                "You have not properly assigned anything to the variable " "{variable}."
            ).format(variable=variable_name),
            label="no_assign",
            title="No Proper Assignment",
            report=report, muted=muted
        )
    else:
        explain(
            ("You have not assigned a {type} to the variable {variable}." "").format(
                type=type, variable=variable_name
            ),
            label="type_assign",
            title="Unexpected Variable Type",
            report=report, muted=muted
        )
    return False



"""

TODO: Handle sys.argv
"""

from pprint import pprint
import ast
import re
import sys
import io
import os
import string
from unittest.mock import patch

from pedal.core.report import MAIN_REPORT
from pedal.core.commands import feedback
from pedal.sandbox.feedbacks import runtime_error
from pedal.utilities.exceptions import ExpandedTraceback, add_context_to_error
from pedal.sandbox import mocked
from pedal.sandbox.exceptions import (SandboxHasNoFunction,
                                      SandboxStudentCodeException,
                                      SandboxHasNoVariable)
from pedal.sandbox.timeout import timeout
from pedal.sandbox.messages import EXTENDED_ERROR_EXPLANATION
from pedal.sandbox.result import SandboxResult
from pedal.sandbox.tracer import (SandboxCallTracer, SandboxCoverageTracer,
                                  SandboxBasicTracer)


def _dict_extends(d1, d2):
    """
    Helper function to create a new dictionary with the contents of the two
    given dictionaries. Does not modify either dictionary, and the values are
    copied shallowly. If there are repeats, the second dictionary wins ties.

    The function is written to ensure Skulpt compatibility.

    Args:
        d1 (dict): The first dictionary
        d2 (dict): The second dictionary
    Returns:
        dict: The new dictionary
    """
    d3 = {}
    for key, value in d1.items():
        d3[key] = value
    for key, value in d2.items():
        d3[key] = value
    return d3


class SandboxVariable:
    """

    """
    def __init__(self, name, value):
        self.name = name
        self.value = value


class DataSandbox:
    """
    Simplistic Mixin class that contains the functions for accessing a
    self-contained student data namespace.
    """

    def __init__(self):
        super().__init__()
        self.data = {}

    def get_names_by_type(self, type, exclude_builtins=True):
        """

        Args:
            type:
            exclude_builtins:

        Returns:

        """
        result = []
        for name, value in self.data.items():
            if isinstance(value, type):
                if exclude_builtins and name.startswith('__'):
                    continue
                result.append(name)
        return result

    def get_values_by_type(self, type, exclude_builtins=True):
        """

        Args:
            type:
            exclude_builtins:

        Returns:

        """
        names = self.get_names_by_type(type, exclude_builtins)
        return [self.data[name] for name in names]

    def get_variables_by_type(self, type, exclude_builtins=True):
        """

        Args:
            type:
            exclude_builtins:

        Returns:

        """
        names = self.get_names_by_type(type, exclude_builtins)
        return [(name, self.data[name]) for name in names]

    @property
    def functions(self):
        """
        Retrieve a list of all the callable names in the students' namespace.
        In other words, get a list of all the functions the student defined.

        Returns:
            list of callables
        """
        return {k: v for k, v in self.data.items() if callable(v)}

    @property
    def var(self):
        """

        Returns:

        """
        return {k: SandboxVariable(k, v) for k, v in self.data.items()}

    def __repr__(self):
        return ""


class Sandbox(DataSandbox):
    """

    The Sandbox is a container that can safely execute student code and store
    the result.

    Attributes:
        data: The namespace produced by the students' code. This is basically
            a dictionary mapping valid python names to their values.
        raw_output (str): The exact literal results of all the `print` calls
            made so far, including the "\n" characters.
        output (list of str): The current lines of output, broken up by
            distinct print calls (not "\n" characters). Note that this will
            not have any "\n" characters unless you explicitly printed them.
        output_contexts (dict[str:list[str]]): The output for each call context.
        call_id (int): The current call_id of the most recent call. Is
            initially 0, indicating the original sandbox creation.
        modules: A dictionary of the mocked modules (accessible by their
            imported names).
        context: A list of strings representing the code previously run through
            this sandbox via .call.
        contextualize (bool): Whether or not to contextualize stack frames.
    """

    CONTEXT_MESSAGE = (
        "\n\nThe error above occurred when I ran:<br>\n<pre>{context}</pre>"
    )
    FILE_CONTEXT_MESSAGE = (
        "\n\nThe error above occurred when I ran your file: {filename}"
    )
    INPUT_CONTEXT_MESSAGE = (
        "And entered the inputs:\n```\n{inputs}\n```"
    )
    TRACER_STYLES = {
        'coverage': SandboxCoverageTracer,
        'calls': SandboxCallTracer,
        'none': SandboxBasicTracer,
    }

    def __init__(self, initial_data=None,
                 initial_raw_output=None,
                 initial_exception=None,
                 modules=None, full_traceback=False,
                 tracer_style='none',
                 threaded=False, report=None,
                 context=None, result_proxy=SandboxResult,
                 instructor_filename="instructor_tests.py",
                 allowed_functions=None):
        """
        Args:
            initial_data (dict[str:Any]): An initial namespace to provide when
                executing the students' code. The keys must be strings and
                should be valid Python names. Defaults to None, which will be
                an empty namespace.
            initial_exception (Exception): An initial exception to load into
                the Sandbox. Usually you will let the students' code generate
                its own exceptions, but if you're constructing a sandbox you
                might need to specify one. Defaults to None.
            modules: A dictionary of strings (valid python package names) that
                map to either the value True (if we provide a default
                implementation) or a user-created MockedModule. By default,
                we mock out the following modules:
                * matplotlib
                * pedal
            context (False, None, or list[str]): How to contextualize calls by
                default in this Sandbox. False means no contextualization.
                None (default) means contextualize automatically. If you give
                a list[str], then it assumes you want to contextualize
                automatically but starting off with the given strings.
            initial_raw_output (str): The initial printed output for the
                sandbox. Usually defaults to None to indicate a blank printed
                area.
            instructor_filename (str): The filename to display in tracebacks,
                when executing student code in instructor tests. Although you
                can specify something else, defaults to "instructor_tests.py".
        """
        super().__init__()
        if initial_data is None:
            initial_data = {}
        self.data = initial_data

        # Context
        self.call_id = 0
        self.target_contexts = {self.call_id: []}
        self.call_contexts = {self.call_id: []}
        self.input_contexts = {self.call_id: []}
        self.context = context
        self.keep_context = False
        # Update outputs
        self.set_output(initial_raw_output)
        # filename
        self.instructor_filename = instructor_filename
        # Temporary data
        self._temporaries = set()
        self._backups = {}
        # Exception
        self.exception = initial_exception
        self.exception_position = None
        self.exception_formatted = None
        self.report_exceptions_mode = False
        self.raise_exceptions_mode = False
        # Input
        self.set_input(None)
        self._input_tracker = self._track_inputs()
        # Modules
        if modules is None:
            modules = {'matplotlib': True,
                       'pedal': mocked.MockPedal()
                       }
        self.mocked_modules = {}
        self.modules = {}
        self.add_mocks(modules)
        self.mocked_functions = {
            'compile': mocked._disabled_compile,
            'eval': mocked._disabled_eval,
            'exec': mocked._disabled_exec,
            'globals': mocked._disabled_globals,
            'open': mocked._restricted_open,
            '__import__': mocked._restricted_import,
        }
        if allowed_functions is not None:
            for function_name in allowed_functions:
                if function_name in self.mocked_functions:
                    del self.mocked_functions[function_name]
        # Patching
        self._current_patches = []
        # Settings
        self.full_traceback = full_traceback
        self.MAXIMUM_VALUE_LENGTH = 120
        # Tracer Styles
        self.tracer_style = tracer_style
        # Proxying results
        self.result_proxy = result_proxy
        # report
        if report is None:
            report = MAIN_REPORT
        self.report = report
        # Threading
        self.threaded = threaded
        self.allowed_time = 3

    def _set_tracer_style(self, tracer_style):
        self._tracer_style = tracer_style.lower()
        self.trace = self.TRACER_STYLES[tracer_style.lower()]()

    def _get_tracer_style(self):
        return self._tracer_style

    tracer_style = property(_get_tracer_style, _set_tracer_style)

    def add_mocks(self, modules):
        """
        :param modules: Keyword listing of modules and their contents
                        (MockedModules) or True (if its one that we have a
                        default implementation for).
        :type modules: dict
        """
        for module_name, module_data in modules.items():
            self._add_mock(module_name, module_data)

    def _add_mock(self, module_name, module_data):
        # MatPlotLib's PyPlot
        if module_name == 'matplotlib':
            matplotlib, modules = mocked.create_module('matplotlib.pyplot')
            self.mocked_modules.update(modules)
            if module_data is True:
                mock_plt = mocked.MockPlt()
                mock_plt._add_to_module(matplotlib.pyplot)
                self.modules['matplotlib.pyplot'] = mock_plt
            else:
                module_data._add_to_module(matplotlib.pyplot)
        else:
            root, modules = mocked.create_module(module_name)
            self.mocked_modules.update(modules)
            self.modules[module_name] = module_data
            module_data._add_to_module(root)

    def set_output(self, raw_output):
        """
        Change the current printed output for the sandbox to the given value.
        If None is given, then clears all the given output (empty list for
        `output` and empty string for `raw_output`).

        Args:
            raw_output (str): The new raw_output for the sandbox. To compute
                the `output` attribute, the system splits and rstrips at
                newlines.
        """
        if raw_output is None:
            self.raw_output = ""
            self.output = []
            self.output_contexts = {self.call_id: list(self.output)}
        else:
            self.raw_output = raw_output
            lines = raw_output.rstrip().split("\n")
            self.output = [line.rstrip() for line in lines]
            self.output_contexts[self.call_id] = list(self.output)

    def append_output(self, raw_output):
        """
        Adds the string of `raw_output` to the current `raw_output` attribute.
        The added string will be split on newlines and rstripped to append
        to the `output` attribute.

        Args:
            raw_output (str): The new raw_output for the sandbox. To compute
                the `output` attribute, the system splits and rstrips at
                newlines.
        """
        self.raw_output += raw_output
        lines = raw_output.rstrip().split("\n")
        lines = [line.rstrip() for line in lines]
        if self.raw_output:
            self.output.extend(lines)
            self.output_contexts[self.call_id].extend(lines)

    def set_input(self, inputs, clear=True):
        """
        Queues the given value as the next arguments to the `input` function.
        """
        if inputs is None:
            self.inputs = []
        if clear:
            self.inputs.clear()
        if isinstance(inputs, str):
            self.inputs.append(inputs)
        elif isinstance(inputs, (list, tuple)):
            self.inputs.extend(inputs)
        elif inputs is not None:
            # TODO: intelligently handle custom generator
            self.inputs = inputs

    def _track_inputs(self):
        """
        Wraps an input function with a tracker.
        """

        def _input_tracker(*args, **kwargs):
            if args:
                prompt = args[0]
            else:
                prompt = ""
            print(prompt)
            if self.inputs:
                value_entered = self.inputs.pop(0)
            else:
                # TODO: Make this smarter, more elegant in choosing IF we should repeat 0
                value_entered = '0'
            self.input_contexts[self.call_id].append(value_entered)
            return value_entered

        return _input_tracker

    def _purge_temporaries(self):
        """
        Delete any variables in the namespace that have been made as
        temporaries. This happens automatically after you execute code.
        """
        for key in self._temporaries:
            if key in self._backups:
                self.data[key] = self.backups[key]
            else:
                del self.data[key]
        self._temporaries = set()

    def _is_long_value(self, value):
        return len(repr(value)) > 25

    def _make_temporary(self, category, name, value, context):
        """
        Create a temporary variable in the namespace for the given
        category/name. This is used to load arguments into the namespace to
        be used in function calls. Temporaries are only created if the value's
        repr length is too long, as defined by _is_long_value.

        Args:
            category (str): A categorical division for the temporary variable
                that can help keep the namespace distinctive - there are a
                few different kinds of categories (e.g., for regular positional
                args, star args, kwargs).
            name (str): A distinctive ID for this variable. The final variable
                name will be "_temporary_<category>_<name>".
            value: The value for this argument.
        Returns:
            str: The new name for the temporary variable.
        """
        if isinstance(value, SandboxVariable):
            return value.name
        if not self._is_long_value(value):
            return repr(value)
        key = '_temporary_{}_{}'.format(category, name)
        if key in self.data:
            self._backups[key] = self.data[key]
        self._temporaries.add(key)
        self.data[key] = value
        if context is None:
            self.call_contexts[self.call_id].append("{} = {}".format(key, value))
        return key

    def run_file(self, filename, as_filename=None, modules=None, inputs=None,
                 threaded=None, context=None, report_exceptions=None,
                 raise_exceptions=None):
        """
        Load the given filename and execute it within the current namespace.
        
        Args:
            raise_exceptions:
            report_exceptions:
            threaded:
            inputs:
            modules:
            as_filename:
            filename:
            context (False, None, or list[str]): The context to give any
                exceptions. If None, then the recorded context will be used. If
                a string, tracebacks will be shown with the given context. If
                False, no context will be given.
        """
        if as_filename is None:
            as_filename = filename
        with open(filename, 'r') as code_file:
            code = code_file.read() + '\n'
        self.run(code, as_filename, modules, inputs, threaded,
                 context, report_exceptions, raise_exceptions)

    def list(self, *args):
        """

        Args:
            *args:
        """
        pass

    def call(self, function, *args, **kwargs):
        """
        Executes the given function with the specified arguments, from the student's code.

        Args:
            function (str): The name of the function to call that was defined
                by the user.
        Returns:
            If the call was successful, returns the result of executing the
            code. Otherwise, it will return an Exception relevant to the
            failure (might be a SandboxException, might be a user-space
            exception).
        """
        # Confirm that the function_name exists
        if function not in self.functions:
            if function not in self.data:
                self.exception = SandboxHasNoVariable(
                    "The function {function} does not exist.".format(function=function)
                )
            else:
                self.exception = SandboxHasNoFunction(
                    "The variable {function} is not a function.".format(function=function)
                )
            return self.exception
        # Parse kwargs for any special arguments.
        as_filename = kwargs.pop('as_filename', self.instructor_filename)
        target = kwargs.pop('target', '_')
        modules = kwargs.pop('modules', {})
        inputs = kwargs.pop('inputs', None)
        threaded = kwargs.pop('threaded', self.threaded)
        context = kwargs.pop('context', self.context)
        keep_context = kwargs.pop('keep_context', self.keep_context)
        report_exceptions = kwargs.pop('report_exceptions', self.report_exceptions_mode)
        raise_exceptions = kwargs.pop('raise_exceptions', self.raise_exceptions_mode)
        # Create the actual arguments and call
        if not keep_context or not self.call_id:
            self.call_id += 1
            self.output_contexts[self.call_id] = []
            self.call_contexts[self.call_id] = []
            self.input_contexts[self.call_id] = []
        # Always update the target context to be most recent
        self.target_contexts[self.call_id] = target
        actual, student = self._construct_call(function, args, kwargs, target,
                                               context)
        if context is None:
            context = student
        # if context is None:
        # self.call_contexts[self.call_id].append(student_call)
        # if context is not False:
        #    self.call_contexts[self.call_id] = context
        self.run(actual, as_filename=as_filename, modules=modules,
                 inputs=inputs, threaded=threaded,
                 context=context, keep_context=keep_context,
                 report_exceptions=report_exceptions,
                 raise_exceptions=raise_exceptions)
        self._purge_temporaries()
        if self.exception is None:
            self._ = self.data[target]
            if self.result_proxy is not None:
                self._ = self.result_proxy(self._, call_id=self.call_id,
                                           sandbox=self)
            return self._
        else:
            # TODO: Might need to wrap this in case the student was supposed
            # to return an exception - weird circumstance though
            return self.exception

    def make_safe_variable(self, name):
        """
        Tries to construct a safe variable name in the current namespace, based
        off the given one. This is accomplished by appending a "_" and a number
        of increasing value until no comparable name exists in the namespace.
        This is particularly useful when you want to create a variable name to
        assign to, but you are concerned that the user might have a variable
        with that name already, which their code relies on.
        
        Args:
            name (str): A desired target name.
        Returns:
            str: A safe target name, based off the given one.
        """
        current_addition = ""
        attempt_index = 2
        while name + current_addition in self.data:
            current_addition = "_{}".format(attempt_index)
            attempt_index += 1
        return name + current_addition

    def _construct_call(self, function, args, kwargs, target, context):
        str_args = [self._make_temporary('arg', index, value, context)
                    for index, value in enumerate(args)]
        str_kwargs = ["{}={}".format(key,
                                     self._make_temporary('kwarg', key, value, context))
                      for key, value in kwargs.items()]
        arguments = ", ".join(str_args + str_kwargs)
        call = "{}({})".format(function, arguments)
        if target is None:
            actual = call
        else:
            actual = "{} = {}".format(target, call)
        student_call = call if target is "_" else actual
        return actual, student_call

    def _start_patches(self, *patches):
        self._current_patches.append(patches)
        for patch in patches:
            patch.start()

    def _stop_patches(self):
        if not self._current_patches:
            return
        patches = self._current_patches.pop()
        for patch in patches:
            patch.stop()

    def _capture_exception(self, exception, exc_info, report_exceptions,
                           raise_exceptions, context, keep_context,
                           as_filename="", code=""):
        self.exception = exception
        student_filename = as_filename or self.report.submission.main_file
        if context is not False:
            if context is None or keep_context:
                contexts = self.call_contexts[self.call_id]
                if context is not None:
                    contexts.append(context)
                context = '\n'.join(contexts)#[1:])
            if context.strip():
                context = self.CONTEXT_MESSAGE.format(context=context)
                inputs = self.input_contexts[self.call_id]
                if inputs is not None and inputs:
                    inputs = "\n".join(inputs)
                    context += "\n"+self.INPUT_CONTEXT_MESSAGE.format(inputs=inputs)
            else:
                context = self.FILE_CONTEXT_MESSAGE.format(filename=student_filename)
            self.exception = add_context_to_error(self.exception, context)
        line_offset = self.report['source'].get('line_offset', 0)
        lines = code.split("\n") if self.report.submission is None else self.report.submission.lines
        filename = as_filename if self.report.submission is None else self.report.submission.main_file
        traceback = ExpandedTraceback(self.exception, exc_info,
                                      self.full_traceback,
                                      self.instructor_filename,
                                      line_offset, filename,
                                      lines)
        self.exception_position = {'line': traceback.line_number}
        self.exception_formatted = traceback.format_exception()
        self.exception_name = str(self.exception.__class__)[8:-2]
        # Do we add the exception to the report?
        if report_exceptions is False:
            return True
        if report_exceptions is None and not self.report_exceptions_mode:
            return True
        runtime_error(self.exception_formatted, self.exception_name, self.exception, traceback.line_number,
                      group=self.report.group, report=self.report)
        if raise_exceptions is True:
            raise SandboxStudentCodeException(self.exception)
        return False

    def run(self, code, as_filename=None, modules=None, inputs=None,
            threaded=None, report_exceptions=True, raise_exceptions=False,
            context=False, keep_context=False):
        """
        Execute the given string of code in this sandbox.
        
        Args:
            keep_context:
            raise_exceptions:
            code (str): The string of code to be executed.
            as_filename (str): The filename to use when executing the code -
                this is cosmetic, technically speaking, it has no relation
                to anything on disk. It will be present in tracebacks.
                Defaults to Source's filename.
            modules (dict[str:Module]): Modules to mock.
            inputs (list[str]): The inputs to give from STDIN, as a list of
                strings. You can also give a function that emulates the
                input function; e.g., consuming a prompt (str) and producing
                strings. This could be used to make a more interactive input
                system.
            context (str): The context to give any exceptions.
                If None, then the recorded context will be used. If a string,
                tracebacks will be shown with the given context. If False,
                no context will be given (the default).
            threaded (bool): whether or not to run this code in a separate
                thread. Defaults to :attribute:`Sandbox.threaded`.
            report_exceptions (bool): Whether or not to capture exceptions.
        """
        # Handle any threading if necessary
        if threaded is None:
            threaded = self.threaded
        if threaded:
            try:
                return timeout(self.allowed_time, self.run, code, as_filename,
                               modules, inputs, False,
                               report_exceptions, raise_exceptions,
                               context, keep_context)
            except TimeoutError as timeout_exception:
                self._stop_patches()
                self._capture_exception(timeout_exception, sys.exc_info(),
                                        report_exceptions, raise_exceptions,
                                        context, keep_context, as_filename,
                                        code)
                return self
        
        if as_filename is None:
            as_filename = os.path.basename(self.report.submission.main_file)
            # todo: earlier version of inputs being made?
        if inputs is not None:
            self.set_input(inputs)
        # Override builtins and mock stuff out
        mocked_functions = self.mocked_functions.copy()
        mocked_functions['input'] = self._input_tracker
        mocked_functions['raw_input'] = self._input_tracker
        mocked_functions['sys'] = sys
        mocked_functions['os'] = os
        mocked._override_builtins(self.data, mocked_functions)

        self.exception = None
        self.exception_position = None
        self.exception_formatted = None

        # Patch in dangerous built-ins
        x = sys.stdout
        capture_stdout = io.StringIO()
        self._start_patches(
            patch.dict('sys.modules', self.mocked_modules),
            patch('sys.stdout', capture_stdout),
            patch('time.sleep', return_value=None),
        )
        # TODO: Hack, add more flexibile way to specify unusable modules
        for module in list(sys.modules.keys()):
            if module.startswith('pedal.'):
                del sys.modules[module]
        try:
            compiled_code = compile(code, as_filename, 'exec')
            with self.trace._as_filename(as_filename, code):
                exec(compiled_code, self.data)
        except Exception as user_exception:
            self._stop_patches()
            info = sys.exc_info()
            self._capture_exception(user_exception, info,
                                    report_exceptions, raise_exceptions,
                                    context, keep_context, as_filename,
                                    code)
        else:
            self._stop_patches()
        finally:
            self.append_output(capture_stdout.getvalue())
        if context is None:
            self.call_contexts[self.call_id].append(code)
        elif isinstance(context, str):
            self.call_contexts[self.call_id].append(context)
        elif context is not False:
            self.call_contexts[self.call_id] = context
        return self


def run(initial_data=None, initial_raw_output=None, initial_exception=None,
        allowed_functions=None,
        modules=None, inputs=None, report_exceptions=True, raise_exceptions=False,
        context=None,
        full_traceback=False, tracer_style='none', threaded=False,
        result_proxy=SandboxResult,
        instructor_filename="instructor_tests.py",
        code=None, as_filename=None, report=None):
    """

    Args:
        initial_data:
        initial_raw_output:
        initial_exception:
        allowed_functions:
        modules:
        inputs:
        report_exceptions:
        raise_exceptions:
        context:
        full_traceback:
        tracer_style:
        threaded:
        result_proxy:
        instructor_filename:
        code:
        as_filename:
        report:

    Returns:

    """
    if report is None:
        report = MAIN_REPORT
    if 'run' not in report['sandbox']:
        report['sandbox']['settings'] = [
            initial_data, initial_raw_output, initial_exception, modules,
            full_traceback, tracer_style, threaded, report, context,
            result_proxy, instructor_filename, allowed_functions
        ]
        report['sandbox']['run'] = Sandbox(*report['sandbox']['settings'])

    sandbox = report['sandbox']['run']
    if code is None:
        code = report.submission.main_code
    sandbox.run(code, as_filename, modules, inputs, threaded,
                report_exceptions, raise_exceptions, context=context, keep_context=False)
    return sandbox


def reset(report=None):
    """

    Args:
        report:
    """
    if report is None:
        report = MAIN_REPORT
    if 'settings' in report['sandbox']:
        report['sandbox']['run'] = Sandbox(*report['sandbox']['settings'])
    else:
        run(report=report)



"""
A package of tools for capturing student code by matching it against patterns.

Requires:
- :py:class:`pedal.source.source.Source`

Optional:
- :py:class:`pedal.tifa.tifa.Tifa`
"""

NAME = 'CAIT'
SHORT_DESCRIPTION = "Captures instructor code patterns within student code."
DESCRIPTION = '''
'''
REQUIRES = ['Source']
OPTIONALS = ['TIFA']

from pedal.cait.cait_api import (find_match, find_matches,
                                 parse_program,
                                 find_submatches, find_expr_sub_matches,
                                 def_use_error, data_state, data_type,
                                 expire_cait_cache)



"""
Imperative style commands for constructing feedback in a convenient way.
Uses a global report object (MAIN_REPORT).
"""

__all__ = ['feedback', 'set_success', 'compliment', 'give_partial', 'explain',
           'gently', 'hide_correctness', 'suppress', 'log', 'debug', 'system_error',
           'clear_report', 'get_all_feedback', 'guidance', 'contextualize_report']

from typing import List

from pedal.core.feedback import Feedback, FeedbackKind, FeedbackCategory, PEDAL_DEVELOPERS, AtomicFeedbackFunction
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT, Report

from pedal.core.submission import Submission

#: Lowercase "function" version that works like other Core Feedback Functions.
feedback = Feedback


# TODO: force_success function, which ignores errors to give the points.


@AtomicFeedbackFunction(title="Complete",
                        text_template="Great work!")
def set_success(score: float = 1, justification: str = None, tool: str = None, group=None,
                report: Report = MAIN_REPORT):
    """
    **(Feedback Function)**

    Creates Successful feedback for the user, indicating that the entire
    assignment is done.

    Args:
        tool:
        group:
        report:
        justification (str): Internal message that explains why this function was used.
        score (number): Arbitrary value to set the user's score to. Defaults to 1.
    """
    return feedback("set_success",
                    tool=tool, category=FeedbackCategory.COMPLETE, kind=FeedbackKind.RESULT,
                    justification=justification, valence=Feedback.POSITIVE_VALENCE,
                    title=set_success.title, message=set_success.text_template.format(),
                    text=set_success.text_template.format(),
                    score=score, correct=True,
                    muted=False, version='1.0.0', author=PEDAL_DEVELOPERS, group=group,
                    report=report)


@AtomicFeedbackFunction(title="Compliment")
def compliment(message: str, value: float = 0, title=None,
               justification: str = None, locations: Location = None, tool: str = None, group=None,
               report: Report = MAIN_REPORT):
    """
    Create a positive feedback for the user, potentially on a specific line of
    code.

    Args:
        report:
        group:
        tool:
        title (str): If None, defaults to "Compliment"
        message (str): The message to display to the user.
        locations (int): The relevant line of code to reference.
        value (float): The number to increase the user's score by.
        justification (str): A justification for why this partial credit was given.
    """
    return feedback("compliment", tool=tool,
                    category=FeedbackCategory.INSTRUCTOR, kind=FeedbackKind.ENCOURAGEMENT,
                    justification=justification, locations=locations,
                    valence=Feedback.POSITIVE_VALENCE,
                    title=title or compliment.title, message=message, text=message,
                    score=value, correct=False, muted=False, version='1.0.0',
                    author=PEDAL_DEVELOPERS, group=group, report=report
                    )


@AtomicFeedbackFunction(title="Partial Credit",
                        text_template="Partial credit")
def give_partial(value: float, justification: str = None, title: str = None, tool: str = None, group=None,
                 report: Report = MAIN_REPORT):
    """
    Increases the user's current score by the `value`.

    Args:
        report:
        group:
        tool:
        title:
        value (number): The number to increase the user's score by.
        justification (str): A justification for why this partial credit was given.
    """
    return feedback("give_partial", tool=tool, category=FeedbackCategory.INSTRUCTOR,
                    kind=FeedbackKind.RESULT,
                    justification=justification, valence=Feedback.POSITIVE_VALENCE,
                    title=title or give_partial.title, message=give_partial.text_template.format(),
                    text=give_partial.text_template.format(),
                    score=value, correct=False, muted=True, version='1.0.0', author=PEDAL_DEVELOPERS, group=group,
                    report=report)


@AtomicFeedbackFunction(title="Instructor Feedback")
def explain(message, label='explain', title=None, fields=None,
            justification=None, priority=None, line=None, text=None, score=None, muted=False,
            version='0.0.1', author=None, tags=None, group=None, report=MAIN_REPORT):
    """

    Args:
        message:
        label:
        title:
        fields:
        justification:
        priority:
        line:
        text:
        score:
        muted:
        version:
        author:
        tags:
        group:
        report:

    Returns:

    """
    return feedback(label, category=Feedback.CATEGORIES.INSTRUCTOR,
                    kind=FeedbackKind.MISTAKE, justification=justification, priority=priority,
                    valence=Feedback.NEGATIVE_VALENCE, title=title or explain.title,
                    message=message or text, text=text or message,
                    fields=fields or {}, locations=line, score=score, muted=muted, version=version,
                    author=author, tags=tags, group=group or report.group, report=report)


@AtomicFeedbackFunction(title="Instructor Feedback")
def gently(message, label='gently', title=None, fields=None,
           justification=None, priority=Feedback.CATEGORIES.STUDENT, line=None, text=None, score=None, muted=False,
           version='0.0.1', author=None, tags=None, group=None, report=MAIN_REPORT):
    """

    Args:
        message:
        label:
        title:
        fields:
        justification:
        priority:
        line:
        text:
        score:
        muted:
        version:
        author:
        tags:
        group:
        report:

    Returns:

    """
    return feedback(label, category=Feedback.CATEGORIES.INSTRUCTOR,
                    kind=FeedbackKind.MISTAKE, justification=justification, priority=priority,
                    valence=Feedback.NEGATIVE_VALENCE, title=title or gently.title,
                    message=message or text, text=text or message,
                    fields=fields or {}, locations=line, score=score, muted=muted, version=version,
                    author=author, tags=tags, group=group or report.group, report=report)


@AtomicFeedbackFunction(title="Instructor Guidance")
def guidance(message, label="guidance", title=None, fields=None,
             justification=None, priority=None, line=None, text=None, score=None,
             muted=False, version='0.0.1', author=None, tags=None, group=None, report=MAIN_REPORT):
    """

    Args:
        message:
        label:
        title:
        fields:
        justification:
        priority:
        line:
        text:
        score:
        muted:
        version:
        author:
        tags:
        group:
        report:

    Returns:

    """
    return feedback(label, category=Feedback.CATEGORIES.INSTRUCTIONS,
                    kind=FeedbackKind.INSTRUCTIONAL, justification=justification, priority=priority,
                    valence=Feedback.NEUTRAL_VALENCE, title=title or guidance.title,
                    message=message or text, text=text or message,
                    fields=fields or {}, locations=line, score=score, muted=muted, version=version,
                    author=author, tags=tags, group=group or report.group, report=report)


def hide_correctness(report=MAIN_REPORT):
    """

    Args:
        report:
    """
    report.hide_correctness()


def suppress(category=None, label=True, report=MAIN_REPORT):
    """

    Args:
        category:
        label:
        report:
    """
    report.suppress(category, label)


def log(message, report=MAIN_REPORT):
    """

    TODO: Consider making this accept star args, like print would.

    Args:
        message:
        report:

    Returns:

    """
    return feedback("log", category=Feedback.CATEGORIES.SYSTEM, muted=True, text=message,
                    valence=Feedback.NEUTRAL_VALENCE)


def debug(message, report=MAIN_REPORT):
    """

    Args:
        message:
        report:

    Returns:

    """
    return feedback("debug", category=Feedback.CATEGORIES.SYSTEM, muted=True, text=message,
                    valence=Feedback.NEGATIVE_VALENCE, priority='high')


def clear_report(report=MAIN_REPORT):
    """
    Removes all existing data from the report, including any submissions, suppressions, feedback,
    and Tool data.

    Args:
        report: The report to clear (defaults to the :py:data:`pedal.core.report.MAIN_REPORT`).
    """
    report.clear()


def contextualize_report(submission, filename='answer.py', clear=True, report=MAIN_REPORT):
    """
    Updates the report with the submission. By default, clears out any old information in the report.
    You can pass in either an actual :py:class:`~pedal.core.submission.Submission` or a string representing
    the code of the submission.

    Args:
        clear:
        submission (str or Submission):
        filename (str or None): If the `submission` was not a :py:class:`~pedal.core.submission.Submission`,
            then this will be used as the filename for the code given in `submission`.
        report: The report to attach this feedback to (defaults to the :py:data:`pedal.core.report.MAIN_REPORT`).
    """
    if not isinstance(submission, Submission):
        submission = Submission(files={filename: submission})
    if clear:
        report.clear()
    report.contextualize(submission)


def get_all_feedback(report=MAIN_REPORT) -> [Feedback]:
    """
    Gives access to the list of feedback from the report. Usually, you won't need this; but if you want
    to build on the results of earlier tools, it can be a useful mechanism.

    TODO: Provide mechanisms for conveniently searching feedback

    Args:
        report: The report to attach this feedback to (defaults to the :py:data:`pedal.core.report.MAIN_REPORT`).

    Returns:
        List[pedal.core.feedback.Feedback]: A list of feedback objects.
    """
    return report.feedback

@AtomicFeedbackFunction(title="System Error")
def system_error(tool: str, explanation: str, author: str = PEDAL_DEVELOPERS, report: Report = MAIN_REPORT):
    """
    Call this function to indicate that something has gone wrong at the system level with Pedal.
    Ideally, this doesn't happen, but sometimes errors cascade and its polite for tools to suggest
    that they are not working correctly. These will not usually be reported to the student.

    Args:
        tool:
        explanation:
        author:
        report:

    Returns:

    """
    return feedback("system_error", tool=tool, category=Feedback.CATEGORIES.SYSTEM,
                    kind=FeedbackKind.META, justification=explanation,
                    valence=Feedback.NEUTRAL_VALENCE,
                    title=system_error.title, message=explanation, text=explanation,
                    muted=True, author=author, report=report)



"""
A package for analyzing student code.

TODO: Statistics
TODO: Output checking for folder of submissions
TODO: Handle tabular output for unit tests
TODO: Big annotated example
TODO: Fix sections example
TODO: Set up "Unit Test" environment
TODO: Add documentation for each one, particularly toolkit
TODO: Set up command line and other environments
TODO: Improve feedback tracking for all the tools
TODO: Fix up imports everywhere to be consistent, particularly for the modules themselves
TODO: A/B Tool - override existing feedback, and tools for randomly or deterministically choosing students

DONE: Rework Source module
DONE: Rework MAIN_REPORT keyword arg for all tools
"""

# Core Commands
from pedal.core.commands import *
from pedal.source import *



"""
Simple data classes for storing feedback to present to learners.
"""

__all__ = ['Feedback', 'FeedbackKind', 'FeedbackCategory', "AtomicFeedbackFunction"]

from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback_category import FeedbackKind, FeedbackCategory


class Feedback:
    """
    A class for storing raw feedback.

    Attributes:
        label (str): An internal name for this specific piece of feedback. The label
            should be an underscore-separated string following the same conventions as
            names in Python. They do not have to be globally unique, but labels should be
            as unique as possible (especially within a category).
        tool (str, optional): An internal name for indicating the tool that created
            this feedback. Should be taken directly from the Tool itself. If `None`, then
            this was not created by a tool but directly by the control script.
        category (str): A human-presentable name showable to the learner, indicating what
            sort of feedback this falls into (e.g., "runtime", "syntax", "algorithm").
            More than one feedback will be in a category, but a feedback cannot be in
            more than one category.
        kind (str): The pedagogical role of this feedback, e.g., "misconception", "mistake",
            "hint", "constraint". Usually, a piece of Feedback is pointing out a mistake,
            but feedback can also be used for various other purposes.
        justification (str): An instructor-facing string briefly describing why this
            feedback was selected. Serves as a "TL;DR" for this feedback category, useful
            for debugging why a piece of feedback appeared.
        priority (str): An indication of how important this feedback is relative to other types
            of feedback in the same category. Might be "high/medium/low". Exactly how this
            gets used is up to the resolver, but typicaly it helps break ties.
        valence (int): Indicates whether this is negative, positive, or neutral feedback.
            Either 1, -1, or 0.

        title (str, optional): A formal, student-facing title for this feedback. If None, indicates
            that the :py:attr:`~pedal.core.feedback.Feedback.label` should be used instead.
        message (str): A markdown-formatted message (aka also supporting HTML) that could be rendered
            to the user.
        text (str): A console-friendly, plain-text message that could be rendered to the user.
        fields (Dict[str,Any]): The raw data that was used to interpolate the template to produce the message.
        locations (:obj:`list` of :py:attr:`~pedal.core.location.Location`): Information about specific locations
            relevant to this message.

        score (int): A numeric score to modify the students' total score, indicating their overall performance.
            It is ultimately up to the Resolver to decide how to combine all the different scores; a typical
            strategy would be to add all the scores together for any non-muted feedback.
        correct (bool): Indicates that the entire submission should be considered correct (success) and that the
            task is now finished.
        muted (bool): Whether this piece of feedback is something that should be shown to a student. There are
            various use cases for muted feedback: they can serve as flags for later conditionals, suppressed
            default kinds of feedback, or perhaps feedback that is interesting for analysis but not pedagogically
            helpful to give to the student. They will still contribute to overall score, but not to the correcntess
            of the submission.

        author (List[str]): A list of names/emails that indicate who created this piece of feedback. They can be
            either names, emails, or combinations in the style of `"Cory Bart <acbart@udel.edu>"`.
        version (str): A version string in the style of Semantic Version (semvar) such as `"0.0.1"`. The last (third)
            digit should be incremented for small bug fixes/changes. The middle (second) digit should be used for more
            serious and intense changes. The first digit should be incremented when changes are made on exposure to
            learners or some other evidence-based motivation.
        tags (List[str]): Any tags that you want to attach to this feedback.

        group (int or str): Information about what logical grouping within the submission this belongs to. Various
            tools can chunk up a submission (e.g., by section), they can use this field to keep track of how that
            decision was made. Resolvers can also use this information to organize feedback or to report multiple
            categories.
    """

    POSITIVE_VALENCE = 1
    NEUTRAL_VALENCE = 0
    NEGATIVE_VALENCE = -1
    CATEGORIES = FeedbackCategory
    KINDS = FeedbackKind

    def __init__(self, label, tool=None, category='instructor', kind='mistake', justification=None,
                 priority=None, valence=-1, title=None, message=None, text=None, fields=None,
                 locations=None, score=None, correct=None, muted=None, version=None, author=None, tags=None,
                 group=None, report=MAIN_REPORT):
        # Model
        self.label = label
        self.tool = tool
        self.category = category
        self.kind = kind
        self.justification = justification
        self.priority = priority
        self.valence = valence
        # Presentation
        self.title = title
        self.message = message
        self.text = text
        self.fields = fields
        if isinstance(locations, int):
            locations = [Location(locations)]
        elif isinstance(locations, Location):
            locations = [locations]
        # TODO: Handle tuples (Line, Col) and (Filename, Line, Col), and possibly lists thereof
        self.locations = locations
        # Result
        self.score = score
        self.correct = correct
        self.muted = muted
        # Metadata
        self.version = version
        self.author = author
        self.tags = tags
        # Organizational
        self.group = group
        # Self-attach to a given report?
        if report is not None:
            report.add_feedback(self)

    def __str__(self):
        """
        Create a string representation of this piece of Feedback for quick, dev purposes.
        Returns:
            str: String representation
        """
        return "<Feedback ({},{})>".format(self.category, self.label)

    def __repr__(self):
        """
        Create a string representation of this piece of Feedback that displays considerably more information.
        Returns:
            str: String representation with more data
        """
        metadata = ""
        if self.tool is not None:
            metadata += ", tool=" + self.tool
        if self.category is not None:
            metadata += ", category=" + self.category
        if self.priority is not None:
            metadata += ", priority=" + self.priority
        if self.group is not None:
            metadata += ", group=" + str(self.group)
        return "Feedback({}{})".format(self.label, metadata)


def AtomicFeedbackFunction(title=None, version='0.0.1', message_template=None, text_template=None,
                           justification=None, tags=None, muted=None):
    """
    Decorator for feedback functions to indicate their intent.

    Args:
        tags:
        title:
        text_template:
        message_template:
        version:
        justification:

    Returns:

    """
    def AtomicFeedbackFunction_with_attrs(function):
        """

        Args:
            function:

        Returns:

        """
        function.title = title if title is not None else function.__name__
        function.tags = tags if tags is not None else []
        function.version = version
        function.justification = justification
        function.message_template = message_template or text_template
        function.text_template = text_template or message_template
        function.muted = muted
        return function
    return AtomicFeedbackFunction_with_attrs


def CompositeFeedbackFunction():
    """

    Returns:

    """
    def CompositeFeedbackFunction_with_attrs(function):
        """

        Args:
            function:

        Returns:

        """
        return function
    return CompositeFeedbackFunction_with_attrs


PEDAL_DEVELOPERS = ["Austin Cory Bart <acbart@udel.edu>", "Luke Gusukuma <lukesg08@vt.edu>"]



import sys

from pedal.sandbox import TOOL_NAME
from pedal.sandbox.feedbacks import runtime_error
from pedal.sandbox.sandbox import Sandbox
from pedal.sandbox.messages import EXTENDED_ERROR_EXPLANATION

from pedal.core.report import MAIN_REPORT


def run_student(raise_exceptions=False, report=MAIN_REPORT, old_style_messages=False):
    """

    Args:
        raise_exceptions:
        report:
        old_style_messages:

    Returns:

    """
    sandbox = report[TOOL_NAME]['run']
    source_code = report.submission.main_code
    filename = report.submission.main_file
    sandbox.run(source_code, as_filename=filename, report_exceptions=not raise_exceptions)
    if raise_exceptions:
        raise_exception(sandbox.exception, sandbox.exception_position,
                        report=report, message=None if old_style_messages else sandbox.exception_formatted)
    return sandbox.exception


def queue_input(*inputs, **kwargs):
    """

    Args:
        *inputs:
        **kwargs:
    """
    if 'report' not in kwargs:
        report = MAIN_REPORT
    else:
        report = kwargs['report']
    sandbox = report[TOOL_NAME]['run']
    sandbox.set_input(inputs)


def reset_output(report=MAIN_REPORT):
    """

    Args:
        report:
    """
    sandbox = report[TOOL_NAME]['run']
    sandbox.set_output(None)


def get_output(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    sandbox = report[TOOL_NAME]['run']
    return sandbox.output


def get_plots(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    sandbox = report[TOOL_NAME]['run']
    if 'matplotlib.pyplot' in sandbox.modules:
        mock_plt = sandbox.modules['matplotlib.pyplot']
        if hasattr(mock_plt, 'plots'):
            return mock_plt.plots
    return []


def capture_output(function, *args, **kwargs):
    """

    Args:
        function:
        *args:
        **kwargs:

    Returns:

    """
    if 'report' in kwargs:
        report = kwargs['report']
    else:
        report = MAIN_REPORT
    sandbox = report[TOOL_NAME]['run']
    sandbox.set_output(None)
    sandbox.call(function.__name__, *args)
    return sandbox.output


def get_sandbox(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    sandbox = report[TOOL_NAME]['run']
    return sandbox


def raise_exception(exception, position=None, report=MAIN_REPORT, message=None):
    """

    Args:
        exception:
        position:
        report:
        message:

    Returns:

    """
    sandbox = report[TOOL_NAME]['run']
    if exception is None:
        return
    extended = EXTENDED_ERROR_EXPLANATION.get(exception.__class__, "")
    if message is None:
        message = "<pre>{}</pre>\n{}".format(str(exception), extended)
    # Skulpt compatible name lookup
    name = str(exception.__class__)[8:-2]
    runtime_error(message, name, exception, position, report=report)
    sandbox.exception = exception


def get_student_data(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    return get_sandbox(report)


def set_sandbox(sandbox, report=MAIN_REPORT):
    """
    Update the sandbox to hold the new sandbox instance. Particularly useful
    for Skulpt, which needs to set the sandbox in an unusual way.
    """
    report['sandbox']['run'] = sandbox
    return sandbox


def trace_lines(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    sandbox = report[TOOL_NAME]['run']
    if sandbox.tracer_style == 'coverage':
        return sandbox.trace.lines - sandbox.trace.missing
    else:
        return []



class Tag:
    """

    """
    def __init__(self, kind, name, explanation, weight):
        self.kind = kind
        self.name = name
        self.explanation = explanation
        self.weight = weight



from pedal.core.report import MAIN_REPORT, Report
from pedal.sandbox.constants import TOOL_NAME
from pedal.sandbox.feedbacks import runtime_error
from pedal.sandbox.sandbox import Sandbox, DataSandbox

# Compatibility API
'''
run_student
queue_input
reset_output
get_output
'''


def reset(report=MAIN_REPORT):
    """

    Args:
        report:
    """
    report[TOOL_NAME] = {
        'run': Sandbox()
    }


Report.register_tool(TOOL_NAME, reset)


def run(raise_exceptions=True, report=MAIN_REPORT, coverage=False, threaded=False, inputs=None) -> Sandbox:
    """

    Args:
        raise_exceptions:
        report:
        coverage:
        threaded:
        inputs:

    Returns:

    """
    sandbox = report[TOOL_NAME]['run']
    sandbox.threaded = threaded
    source_code = report.submission.main_code
    sandbox.record_coverage = coverage
    sandbox.run(source_code, as_filename=report.submission.main_file, inputs=inputs, threaded=threaded)
    sandbox.raise_exceptions_mode = raise_exceptions
    if raise_exceptions and sandbox.exception is not None:
        name = str(sandbox.exception.__class__)[8:-2]

        runtime_error(sandbox.exception_formatted, name, sandbox.exception, sandbox.exception_position,
                      group=report['source']['section'], report=report)

    return sandbox



from pedal.tifa.type_definitions import (UnknownType, FunctionType,
                                         NumType, NoneType, BoolType,
                                         TupleType, ListType, StrType,
                                         FileType, DictType, ModuleType,
                                         SetType, DayType, TimeType,
                                         LiteralNum)

def get_builtin_module(name):
    """

    Args:
        name:

    Returns:

    """
    if name == 'matplotlib':
        return ModuleType('matplotlib',
                          submodules={
                              'pyplot': ModuleType('pyplot', fields={
                                  'plot': FunctionType(name='plot', returns=NoneType()),
                                  'hist': FunctionType(name='hist', returns=NoneType()),
                                  'scatter': FunctionType(name='scatter', returns=NoneType()),
                                  'show': FunctionType(name='show', returns=NoneType()),
                                  'xlabel': FunctionType(name='xlabel', returns=NoneType()),
                                  'ylabel': FunctionType(name='ylabel', returns=NoneType()),
                                  'title': FunctionType(name='title', returns=NoneType()),
                              })
                          })
    elif name == 'pprint':
        return ModuleType('pprint',
                          fields={
                              'pprint': FunctionType(name='pprint', returns=NoneType())
                          })
    elif name == 'random':
        return ModuleType('random',
                          fields={
                              'randint': FunctionType(name='randint', returns=NumType())
                          })
    elif name == 'string':
        return ModuleType('string',
                          fields={
                              'letters': StrType(empty=False),
                              'digits': StrType(empty=False),
                              'ascii_letters': StrType(empty=False),
                              'punctuation': StrType(empty=False),
                              'printable': StrType(empty=False),
                              'whitespace': StrType(empty=False),
                              'ascii_uppercase': StrType(empty=False),
                              'ascii_lowercase': StrType(empty=False),
                              'hexdigits': StrType(empty=False),
                              'octdigits': StrType(empty=False),
                          })
    elif name == 'turtle':
        return ModuleType('turtle',
                          fields={
                              'forward': FunctionType(name='forward', returns=NoneType()),
                              'backward': FunctionType(name='backward', returns=NoneType()),
                              'color': FunctionType(name='color', returns=NoneType()),
                              'right': FunctionType(name='right', returns=NoneType()),
                              'left': FunctionType(name='left', returns=NoneType()),
                          })
    elif name == 'parking':
        return ModuleType('parking',
                          fields={
                              'Time': FunctionType(name='Time', returns=TimeType()),
                              'now': FunctionType(name='now', returns=TimeType()),
                              'Day': FunctionType(name='Day', returns=DayType()),
                              'today': FunctionType(name='today', returns=DayType()),
                          }),
    elif name == 'math':
        return ModuleType('math',
                          fields={
                              'ceil': FunctionType(name='ceil', returns=NumType()),
                              'copysign': FunctionType(name='copysign', returns=NumType()),
                              'fabs': FunctionType(name='fabs', returns=NumType()),
                              'factorial': FunctionType(name='factorial', returns=NumType()),
                              'floor': FunctionType(name='floor', returns=NumType()),
                              'fmod': FunctionType(name='fmod', returns=NumType()),
                              'frexp': FunctionType(name='frexp', returns=NumType()),
                              'fsum': FunctionType(name='fsum', returns=NumType()),
                              'gcd': FunctionType(name='gcd', returns=NumType()),
                              'isclose': FunctionType(name='isclose', returns=BoolType()),
                              'isfinite': FunctionType(name='isfinite', returns=BoolType()),
                              'isinf': FunctionType(name='isinf', returns=BoolType()),
                              'isnan': FunctionType(name='isnan', returns=BoolType()),
                              'ldexp': FunctionType(name='ldexp', returns=NumType()),
                              'modf': FunctionType(name='modf', returns=NumType()),
                              'trunc': FunctionType(name='trunc', returns=NumType()),
                              'log': FunctionType(name='log', returns=NumType()),
                              'log1p': FunctionType(name='log1p', returns=NumType()),
                              'log2': FunctionType(name='log2', returns=NumType()),
                              'log10': FunctionType(name='log10', returns=NumType()),
                              'pow': FunctionType(name='pow', returns=NumType()),
                              'sqrt': FunctionType(name='sqrt', returns=NumType()),
                              'sin': FunctionType(name='sin', returns=NumType()),
                              'cos': FunctionType(name='cos', returns=NumType()),
                              'tan': FunctionType(name='tan', returns=NumType()),
                              'asin': FunctionType(name='asin', returns=NumType()),
                              'acos': FunctionType(name='acos', returns=NumType()),
                              'atan': FunctionType(name='atan', returns=NumType()),
                              'atan2': FunctionType(name='atan2', returns=NumType()),
                              'hypot': FunctionType(name='hypot', returns=NumType()),
                              'degrees': FunctionType(name='degrees', returns=NumType()),
                              'radians': FunctionType(name='radians', returns=NumType()),
                              'sinh': FunctionType(name='sinh', returns=NumType()),
                              'cosh': FunctionType(name='cosh', returns=NumType()),
                              'tanh': FunctionType(name='tanh', returns=NumType()),
                              'asinh': FunctionType(name='asinh', returns=NumType()),
                              'acosh': FunctionType(name='acosh', returns=NumType()),
                              'atanh': FunctionType(name='atanh', returns=NumType()),
                              'erf': FunctionType(name='erf', returns=NumType()),
                              'erfc': FunctionType(name='erfc', returns=NumType()),
                              'gamma': FunctionType(name='gamma', returns=NumType()),
                              'lgamma': FunctionType(name='lgamma', returns=NumType()),
                              'pi': NumType(),
                              'e': NumType(),
                              'tau': NumType(),
                              'inf': NumType(),
                              'nan': NumType(),
                          })


def _builtin_sequence_constructor(sequence_type):
    """
    Helper function for creating constructors for the Set and List types.
    These constructors use the subtype of the arguments.

    Args:
        sequence_type (Type): A function for creating new sequence types.
    """

    def sequence_call(tifa, function_type, callee, args, position):
        """

        Args:
            tifa:
            function_type:
            callee:
            args:
            position:

        Returns:

        """
        # TODO: Should inherit the emptiness too
        return_type = sequence_type(empty=True)
        if args:
            return_type.subtype = args[0].index(LiteralNum(0))
            return_type.empty = False
        return return_type

    return sequence_call


def _builtin_zip(tifa, function_type, callee, args, position):
    """
    Definition of the built-in zip function, which consumes a series of
    sequences and returns a list of tuples, with each tuple composed of the
    elements of the sequence paired (or rather, tupled) together.
    """
    if args:
        tupled_types = TupleType(subtypes=[])
        for arg in args:
            tupled_types.subtypes.append(arg.index(0))
        return ListType(tupled_types, empty=False)
    return ListType(empty=True)


# TODO: Exceptions

def get_builtin_function(name):
    """

    Args:
        name:

    Returns:

    """
    # Void Functions
    if name == "print":
        return FunctionType(name="print", returns=NoneType())
    # Math Functions
    elif name in ("int", "abs", "float", "len", "ord", "pow", "round", "sum"):
        return FunctionType(name=name, returns=NumType())
    # Boolean Functions
    elif name in ("bool", "all", "any", "isinstance"):
        return FunctionType(name=name, returns=BoolType())
    # String Functions
    elif name in ("str", 'chr', 'bin', 'repr', 'input'):
        return FunctionType(name=name, returns=StrType())
    # File Functions
    elif name == "open":
        return FunctionType(name="open", returns=FileType())
    # List Functions
    elif name == "map":
        return FunctionType(name="map", returns=ListType(empty=False))
    elif name == "list":
        return FunctionType(name="list",
                            definition=_builtin_sequence_constructor(ListType))
    # Set Functions
    elif name == "set":
        return FunctionType(name="set",
                            definition=_builtin_sequence_constructor(SetType))
    # Dict Functions
    elif name == "dict":
        return FunctionType(name="dict", returns=DictType())
    # Pass through
    elif name == "sorted":
        return FunctionType(name="sorted", returns='identity')
    elif name == "reversed":
        return FunctionType(name="reversed", returns='identity')
    elif name == "filter":
        return FunctionType(name="filter", returns='identity')
    # Special Functions
    elif name == "type":
        return FunctionType(name="type", returns=UnknownType())
    elif name == "range":
        return FunctionType(name="range", returns=ListType(NumType(), empty=False))
    elif name == "dir":
        return FunctionType(name="dir", returns=ListType(StrType(), empty=False))
    elif name == "max":
        return FunctionType(name="max", returns='element')
    elif name == "min":
        return FunctionType(name="min", returns='element')
    elif name == "zip":
        return FunctionType(name="zip", returns=_builtin_zip)
    elif name == "__import__":
        return FunctionType(name="__import__", returns=ModuleType())
    elif name == "globals":
        return FunctionType(name="globals",
                            returns=DictType(keys=StrType(),
                                             values=UnknownType(),
                                             empty=False))
    elif name in ("classmethod", "staticmethod"):
        return FunctionType(name=name, returns='identity')
    elif name in ("__name__",):
        return StrType()



__all__ = ['Report', 'MAIN_REPORT']

from pedal.core.feedback_category import FeedbackCategory


class Report:
    """
    A class for storing Feedback generated by Tools, along with any auxiliary
    data that the Tool might want to provide for other tools.

    Attributes:
        submission (:py:class:`~pedal.core.submission.Submission`): The contextualized submission information.
        feedback (list of :py:class:`~pedal.core.feedback.Feedback`): The raw feedback generated for
            this Report so far.
        suppressions (list of tuple(str, str)): The categories and labels that have been suppressed so far.
        group (int or str): The label for the current group. Feedback given
            by a Tool will automatically receive the current `group`. This
            is used by the Source tool, for example, in order to group feedback
            by sections.
        group_names (dict[group:str]): A printable, student-facing name for the
            group. When a group needs to be rendered out to the user, this
            will override whatever label was going to be presented instead.
        group_order (sequence or callable or None): The mechanism to use to
            order groups. If a sequence, the order will be inferred based on
            the order of elements in the sequence. If a callable, the callable
            will be used as a key function for `sort`. If `None`, then defaults
            to the natural ordering of the groups. Defaults to `None`.
        hooks (dict[str: list[callable]): A dictionary mapping events to
            a list of callable functions. Tools can register functions on
            hooks to have them executed when the event is triggered by another
            tool. For example, the Assertions tool has hooks on the Source tool
            to trigger assertion resolutions before advancing to next sections.
        _tool_data (dict of str => any): Maps tool names to their data. The
                                       namespace for a tool can be used to
                                       store whatever they want, but will
                                       probably be in a dictionary itself.
    """
    group_order = None

    TOOLS = {}

    def __init__(self):
        """
        Creates a new Report instance.
        """
        self.clear()

    def clear(self):
        """

        """
        self.feedback = []
        self.suppressions = {}
        self.suppressed_labels = []
        self._tool_data = {}
        self.group = None
        self.group_names = {}
        self.hooks = {}
        self.submission = None
        self.result = None

    def contextualize(self, submission):
        """

        Args:
            submission:
        """
        self.submission = submission

    def hide_correctness(self):
        """

        """
        self.suppressions[FeedbackCategory.RESULT] = []

    def add_feedback(self, feedback):
        """
        Attaches the given feedback object to this report.

        Args:
            feedback:

        Returns:
            :py:class:`~pedal.core.feedback.Feedback`: The attached feedback
        """
        self.feedback.append(feedback)
        return feedback

    def suppress(self, category=None, label=True, where=True):
        """
        Args:
            category (str): The category of feedback to suppress.
            label (bool or str): A specific label to match against and suppress.
            where (bool or group): Which group of report to localize the
                suppression to. If instead `True` is passed, the suppression
                occurs in every group globally.
                TODO: Currently, only global suppression is supported.
        """
        if category is None:
            self.suppressed_labels.append(label)
        else:
            category = category.lower()
            if isinstance(label, str):
                label = label.lower()
            if category in FeedbackCategory.ALIASES:
                category = FeedbackCategory.ALIASES[category]
            if category not in self.suppressions:
                self.suppressions[category] = []
            self.suppressions[category].append(label)

    def add_hook(self, event, function):
        """
        Register the `function` to be executed when the given `event` is
        triggered.
        
        Args:
            event (str): An event name. Multiple functions can be triggered for
                the same `event`. The format is as follows: `"namespace.function.extra"`
                The `".extra"` component is optional to add further nuance, but
                the general idea is that you are referring to functions that,
                when called, should trigger other functions to be called first.
                The namespace is typically a tool or module.
            function (callable): A callable function. This function should
                accept a keyword parameter named `report`; this report will be passed
                as as that argument.
        """
        if event not in self.hooks:
            self.hooks[event] = []
        self.hooks[event].append(function)

    def execute_hooks(self, tool, event_name):
        """
        Trigger the functions for all of the associated hooks.
        Hooks will be called with this report as a keyword `report` argument.

        Args:
            tool (str): The name of the tool, to namespace events by.
            event_name (str): The event name (separate words with periods).
        """
        event = tool + '.' + event_name
        if event in self.hooks:
            for function in self.hooks[event]:
                function(report=self)

    def __getitem__(self, tool_name):
        if tool_name not in self._tool_data:
            self.TOOLS[tool_name]['reset'](report=self)
        return self._tool_data[tool_name]

    def __setitem__(self, tool_name, value):
        self._tool_data[tool_name] = value

    def __contains__(self, tool_name):
        return tool_name in self._tool_data

    @classmethod
    def register_tool(cls, tool_name: str, reset_function):
        """
        Identifies that the given Tool should be made available.
        Args:
            tool_name: A unique string identifying this tool.
            reset_function: The function to call to reset the Tool.

        Returns:

        """
        if tool_name in cls.TOOLS:
            # TODO: More sophisticated exceptions
            raise Exception("Tool namespace {} already registered.")
        cls.TOOLS[tool_name] = {
            'reset': reset_function
        }


#: The global Report object. Meant to be used as a default singleton
#: for any tool, so that instructors do not have to create their own Report.
#: Of course, all APIs are expected to work with a given Report, and only
#: default to this Report when no others are given.
#: Ideally, the average instructor will never know this exists.
MAIN_REPORT = Report()




__all__ = ['Submission']


class Submission:
    """
    Simple class for holding information about the student's submission.

    Examples:
        A very simple example of creating a Submission with a single file::

            >>> Submission({'answer.py': "a = 0"})

    Attributes:
        files (`dict` mapping `str` to `str`): Dictionary of filenames mapped to their contents, emulating
            a file system.
        main_file (str): The entry point file that will be considered the main file.
        main_code (str): The actual code to run; if None, then defaults to the code of the main file.
            Useful for tools that want to change the currently active code (e.g., Source's sections) or
            run additional commands (e.g., Sandboxes' call).
        user (dict): Additional information about the user.
        assignment (dict): Additional information about the assignment.
        course (dict): Additional information about the course.
        execution (dict): Additional information about the results of previously executing the students' code.
    """

    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='on_run.py'):
        self.files = files
        self.main_file = main_file
        if main_code is None:
            self.main_code = self.files[self.main_file]
        else:
            self.main_code = main_code
        self._original_main_code = self.main_code
        self.user = user
        self.assignment = assignment
        self.course = course
        self.execution = execution
        self.instructor_file = instructor_file
        self._lines_cache = {}

    @property
    def lines(self):
        """

        Returns:

        """
        if self.main_code not in self._lines_cache:
            self._lines_cache[self.main_code] = self.main_code.split("\n")
        return self._lines_cache[self.main_code]

    def replace_main(self, code: str, file: str = None):
        """
        Substitutes the current main code and filename with the given arguments.
        Args:
            code:
            file:

        Returns:

        """
        self.main_code = code
        if file is not None:
            self.main_file = file



from pedal.core.report import MAIN_REPORT
from pedal.questions.graders import QuestionGrader


class Question:
    """

    """
    def __init__(self, name, instructions, tests, seed=None, report=MAIN_REPORT):
        self.name = name
        self.instructions = instructions
        self.tests = tests
        self.seed = seed
        self.report = report
        self.answered = False

    def answer(self):
        """

        """
        self.answered = True

    def ask(self):
        """

        """
        if isinstance(self.tests, QuestionGrader):
            self.tests._test(self)
        else:
            for test in self.tests:
                test(self)
        if not self.answered:
            show_question(self.instructions, self.report)


def show_question(instructions, report=MAIN_REPORT):
    """

    Args:
        instructions:
        report:
    """
    report.attach('Question', category='Instructions', tool='Questions',
                  group=report.group, priority='instructions', hint=instructions)



from pedal.cait.cait_api import parse_program
from pedal.core.commands import explain


def ensure_imports(*modules):
    """

    Args:
        *modules:

    Returns:

    """
    ast = parse_program()
    for module in modules:
        imports = ast.find_all("Import")
        import_froms = ast.find_all("ImportFrom")
        if not imports and not import_froms:
            explain("You need to import the <code>{}</code> module.".format(module))
            return True
        success = False
        if imports:
            if any(alias._name == module
                   for i in imports
                   for alias in i.names):
                success = True
        if import_froms:
            if any(i.module == module for i in import_froms):
                success = True
        if not success:
            explain("You need to import the <code>{}</code> module.".format(module))
            return True
    return False



from pedal.cait.cait_api import parse_program
from pedal.core.commands import gently
from pedal.core.report import MAIN_REPORT
from pedal.sandbox import compatibility
import ast

from pedal.toolkit.signatures import type_check, parse_type, normalize_type, parse_type_value, test_type_equality

DELTA = 0.001


def all_documented():
    """

    Returns:

    """
    ast = parse_program()
    defs = ast.find_all('FunctionDef') + ast.find_all("ClassDef")
    for a_def in defs:
        if a_def.name == "__init__":
            continue
        if (a_def.body and
                (a_def.body[0].ast_name != "Expr" or
                 a_def.body[0].value.ast_name != "Str")):
            if a_def.ast_name == 'FunctionDef':
                gently("You have an undocumented function: " + a_def.name)
            else:
                gently("You have an undocumented class: " + a_def.name)
            return False
    return True


def get_arg_name(node):
    """

    Args:
        node:

    Returns:

    """
    name = node.id
    if name is None:
        return node.arg
    else:
        return name


def match_function(name, root=None):
    """

    Args:
        name:
        root:

    Returns:

    """
    if root is None:
        ast = parse_program()
    else:
        ast = root
    defs = ast.find_all('FunctionDef')
    for a_def in defs:
        if a_def._name == name:
            return a_def
    return None


def match_signature_muted(name, length, *parameters):
    """

    Args:
        name:
        length:
        *parameters:

    Returns:

    """
    ast = parse_program()
    defs = ast.find_all('FunctionDef')
    for a_def in defs:
        if a_def._name == name:
            found_length = len(a_def.args.args)
            if found_length != length:
                return None
            elif parameters:
                for parameter, arg in zip(parameters, a_def.args.args):
                    arg_name = get_arg_name(arg)
                    if arg_name != parameter:
                        return None
                else:
                    return a_def
            else:
                return a_def
    return None


def find_def_by_name(name, root=None):
    """

    Args:
        name:
        root:

    Returns:

    """
    if root is None:
        root = parse_program()
    defs = root.find_all('FunctionDef')
    for a_def in defs:
        if a_def._name == name:
            return a_def
    return None


def match_parameters(name, *types, returns=None, root=None):
    """

    Args:
        name:
        *types:
        returns:
        root:

    Returns:

    """
    defn = find_def_by_name(name, root)
    if defn:
        for expected, actual in zip(types, defn.args.args):
            if actual.annotation:
                expected = parse_type_value(expected, True)
                actual_type = parse_type(actual.annotation)
                if not test_type_equality(expected, actual_type):
                    gently("Error in definition of function `{}` parameter `{}`. Expected `{}`, "
                             "instead found `{}`.".format(name, actual.arg, expected, actual_type),
                             label="wrong_parameter_type", title="Wrong Parameter Type")
                    return None
        else:
            if returns is not None:
                if not isinstance(returns, str):
                    returns = returns.__name__
                if defn.returns:
                    actual_type = parse_type(defn.returns)
                    if not type_check(returns, actual_type):
                        gently("Error in definition of function `{}` return type. Expected `{}`, "
                                 "instead found {}.".format(name, returns, actual_type),
                                 label="wrong_return_type")
                        return None
                else:
                    gently("Error in definition of function `{}` return type. Expected `{}`, "
                             "but there was no return type specified.".format(name, returns),
                             label="missing_return_type")
                    return None
            return defn


def match_signature(name, length, *parameters):
    """

    Args:
        name:
        length:
        *parameters:

    Returns:

    """
    ast = parse_program()
    defs = ast.find_all('FunctionDef')
    for a_def in defs:
        if a_def._name == name:
            found_length = len(a_def.args.args)
            if found_length < length:
                gently("The function named <code>{}</code> has fewer parameters ({}) "
                         "than expected ({}). ".format(name, found_length, length), label="insuff_args")
            elif found_length > length:
                gently("The function named <code>{}</code> has more parameters ({}) "
                         "than expected ({}). ".format(name, found_length, length), label="excess_args")
            elif parameters:
                for parameter, arg in zip(parameters, a_def.args.args):
                    arg_name = get_arg_name(arg)
                    if arg_name != parameter:
                        gently("Error in definition of <code>{}</code>. Expected a parameter named {}, "
                                 "instead found {}.".format(name, parameter, arg_name), label="name_missing")
                        return None
                else:
                    return a_def
            else:
                return a_def
    else:
        gently("No function named <code>{name}</code> was found.".format(name=name),
                 label="missing_func", title="Missing Function", fields={'name': name})
    return None


TEST_TABLE_HEADER = "<table class='blockpy-feedback-unit table table-sm table-bordered table-hover'>"
TEST_TABLE_OUTPUT = TEST_TABLE_HEADER+(
    "<tr class='table-active'><th></th><th>Arguments</th><th>Expected</th><th>Actual</th></tr>"
)
TEST_TABLE_UNITS = TEST_TABLE_HEADER+(
    "<tr class='table-active'><th></th><th>Arguments</th><th>Returned</th><th>Expected</th></tr>"
)
GREEN_CHECK = "<td class='green-check-mark'>&#10004;</td>"
RED_X = "<td>&#10060;</td>"


def output_test(name, *tests):
    """

    Args:
        name:
        *tests:

    Returns:

    """
    student = compatibility.get_student_data()
    if name in student.data:
        the_function = student.data[name]
        if callable(the_function):
            result = TEST_TABLE_OUTPUT
            success = True
            success_count = 0
            for test in tests:
                inp = test[:-1]
                inputs = ', '.join(["<code>{}</code>".format(repr(i)) for i in inp])
                out = test[-1]
                tip = ""
                if isinstance(out, tuple):
                    tip = out[1]
                    out = out[0]
                message = "<td><code>{}</code></td>" + ("<td><pre>{}</pre></td>" * 2)
                test_out = compatibility.capture_output(the_function, *inp)
                if isinstance(out, str):
                    if len(test_out) < 1:
                        message = message.format(inputs, repr(out), "<i>No output</i>", tip)
                        message = "<tr class=''>" + RED_X + message + "</tr>"
                        if tip:
                            message += "<tr class='table-info'><td colspan=4>" + tip + "</td></tr>"
                        success = False
                    elif len(test_out) > 1:
                        message = message.format(inputs, "\n".join(out), "<i>Too many outputs</i>", tip)
                        message = "<tr class=''>" + RED_X + message + "</tr>"
                        if tip:
                            message += "<tr class='info'><td colspan=4>" + tip + "</td></tr>"
                        success = False
                    elif out not in test_out:
                        message = message.format(inputs, "\n".join(out), "\n".join(test_out), tip)
                        message = "<tr class=''>" + RED_X + message + "</tr>"
                        if tip:
                            message += "<tr class='table-info'><td colspan=4>" + tip + "</td></tr>"
                        success = False
                    else:
                        message = message.format(inputs, "\n".join(out), "\n".join(test_out), tip)
                        message = "<tr class=''>" + GREEN_CHECK + message + "</tr>"
                        success_count += 1
                elif out != test_out:
                    if len(test_out) < 1:
                        message = message.format(inputs, "\n".join(out), "<i>No output</i>", tip)
                    else:
                        message = message.format(inputs, "\n".join(out), "\n".join(test_out), tip)
                    message = "<tr class=''>" + RED_X + message + "</tr>"
                    if tip:
                        message += "<tr class='table-info'><td colspan=4>" + tip + "</td></tr>"
                    success = False
                else:
                    message = message.format(inputs, "\n".join(out), "\n".join(test_out), tip)
                    message = "<tr class=''>" + GREEN_CHECK + message + "</tr>"
                    success_count += 1
                result += message
            if success:
                return the_function
            else:
                result = ("I ran your function <code>{}</code> on some new arguments, and it gave the wrong output "
                          "{}/{} times.".format(name, len(tests) - success_count, len(tests)) + result)
                gently(result + "</table>", label="wrong_output")
                return None
        else:
            gently("You defined {}, but did not define it as a function.".format(name), label="not_func_def")
            return None
    else:
        gently("The function <code>{}</code> was not defined.".format(name), label="no_func_def")
        return None


def unit_test(name, *tests):
    """
    Show a table
    :param name:
    :param tests:
    :return:
    """
    student = compatibility.get_student_data()
    if name in student.data:
        the_function = student.data[name]
        if callable(the_function):
            result = TEST_TABLE_UNITS
            success = True
            success_count = 0
            for test in tests:
                inp = test[:-1]
                inputs = ', '.join(["<code>{}</code>".format(repr(i)) for i in inp])
                out = test[-1]
                tip = ""
                if isinstance(out, tuple):
                    tip = out[1]
                    out = out[0]
                message = ("<td><code>{}</code></td>" * 3)
                ran = True
                try:
                    test_out = the_function(*inp)
                except Exception as e:
                    message = message.format(inputs, str(e), repr(out))
                    message = "<tr class=''>" + RED_X + message + "</tr>"
                    success = False
                    ran = False
                if not ran:
                    result += message
                    continue
                message = message.format(inputs, repr(test_out), repr(out))
                if (isinstance(out, float) and
                        isinstance(test_out, (float, int)) and
                        abs(out - test_out) < DELTA):
                    message = "<tr class=''>" + GREEN_CHECK + message + "</tr>"
                    success_count += 1
                elif out != test_out:
                    # gently(message)
                    message = "<tr class=''>" + RED_X + message + "</tr>"
                    if tip:
                        message += "<tr class='table-info'><td colspan=4>" + tip + "</td></tr>"
                    success = False
                else:
                    message = "<tr class=''>" + GREEN_CHECK + message + "</tr>"
                    success_count += 1
                result += message
            if success:
                return the_function
            else:
                result = "I ran your function <code>{}</code> on some new arguments, " \
                         "and it failed {}/{} tests.".format(name, len(tests) - success_count, len(tests)) + result
                gently(result + "</table>", label="tests_failed")
                return None
        else:
            gently("You defined {}, but did not define it as a function.".format(name))
            return None
    else:
        gently("The function <code>{}</code> was not defined.".format(name))
        return None


class _LineVisitor(ast.NodeVisitor):
    """
    NodeVisitor subclass that visits every statement of a program and tracks
    their line numbers in a list.
    
    Attributes:
        lines (list[int]): The list of lines that were visited.
    """

    def __init__(self):
        self.lines = []

    def _track_lines(self, node):
        self.lines.append(node.lineno)
        self.generic_visit(node)

    visit_FunctionDef = _track_lines
    visit_AsyncFunctionDef = _track_lines
    visit_ClassDef = _track_lines
    visit_Return = _track_lines
    visit_Delete = _track_lines
    visit_Assign = _track_lines
    visit_AugAssign = _track_lines
    visit_AnnAssign = _track_lines
    visit_For = _track_lines
    visit_AsyncFor = _track_lines
    visit_While = _track_lines
    visit_If = _track_lines
    visit_With = _track_lines
    visit_AsyncWith = _track_lines
    visit_Raise = _track_lines
    visit_Try = _track_lines
    visit_Assert = _track_lines
    visit_Import = _track_lines
    visit_ImportFrom = _track_lines
    visit_Global = _track_lines
    visit_Nonlocal = _track_lines
    visit_Expr = _track_lines
    visit_Pass = _track_lines
    visit_Continue = _track_lines
    visit_Break = _track_lines


def check_coverage(report=None):
    """
    Checks that all the statements in the program have been executed.
    This function only works when a tracer_style has been set in the sandbox,
    or you are using an environment that automatically traces calls (e.g.,
    BlockPy).
    
    TODO: Make compatible with tracer_style='coverage'
    
    Args:
        report (Report): The Report to draw source code from; if not given,
            defaults to MAIN_REPORT.
    Returns:
        bool or set[int]: If the source file was not parsed, None is returned.
            If there were fewer lines traced in execution than are found in
            the AST, then the set of unexecuted lines are returned. Otherwise,
            False is returned.
    """
    if report is None:
        report = MAIN_REPORT
    if not report['source']['success']:
        return None, 0
    lines_executed = set(compatibility.trace_lines())
    if -1 in lines_executed:
        lines_executed.remove(-1)
    student_ast = report['source']['ast']
    visitor = _LineVisitor()
    visitor.visit(student_ast)
    lines_in_code = set(visitor.lines)
    if lines_executed < lines_in_code:
        return lines_in_code - lines_executed, len(lines_executed)/len(lines_in_code)
    else:
        return False, 1


def ensure_coverage(percentage=.5, destructive=False, report=None):
    """
    Provides some feedback if the students' code has coverage below the
    given percentage.
    Note that this avoids destroying the current sandbox instance stored on the
    report, if there is one present.
    
    Args:
        report:
        percentage:
        destructive (bool): Whether or not to remove the sandbox.
    """
    if report is None:
        report = MAIN_REPORT
    student_code = report['source']['code']
    unexecuted_lines, percent_covered = check_coverage(report)
    if unexecuted_lines:
        if percent_covered <= percentage:
            gently("Your code coverage is not adequate. You must cover at least half your code to receive feedback.")
            return False
    return True


def ensure_cisc108_tests(test_count, report=None):
    """

    TODO: This should be moved out of pedal, or generalized for environments.

    Args:
        test_count:
        report:

    Returns:

    """
    student = compatibility.get_student_data()
    if 'assert_equal' not in student.data:
        gently("You have not imported assert_equal from the cisc108 module.")
        return False
    assert_equal = student.data['assert_equal']
    if not hasattr(assert_equal, 'student_tests'):
        gently("The assert_equal function has been modified. Do not let it be overwritten!",
               label="Assertion Function Corrupted")
        return False
    student_tests = assert_equal.student_tests
    if student_tests.tests == 0:
        gently("You are not unit testing the result.", label="No Student Unit Tests")
        return False
    elif student_tests.tests < test_count:
        gently("You have not written enough unit tests.", label="Not Enough Student Unit Tests")
        return False
    elif student_tests.failures > 0:
        gently("Your unit tests are not passing.", label="Student Unit Tests Failing")
        return False
    return True



"""
Tools are effectively submodules within Pedal - notable exceptions are :py:mod:`pedal.core`,
:py:mod:`pedal.environments`, :py:mod:`pedal.utilities`, and :py:mod:`pedal.resolvers`.

All Tools with any kind of state are expected to have a `reset` function. Although this can
take parameters, we recommend avoiding that. The `reset` should be for putting things back into
a "null" state, and then you could have followup functions that also give initial state.

Tools should define a `constants.py` file with any useful constants. One of these should be
`TOOL_NAME`, a string value indicating their desired namespace (e.g., `TOOL_NAME = "source").

Tools should define a `feedbacks.py` module that centralizes Atomic Feedback funtions for that tool.

Any function that interacts with Reports should expose a `report` parameter that defaults to
:py:data:`pedal.core.report.MAIN_REPORT`. All internal functions should respect the `report`
that was passed in, and not assume the `MAIN_REPORT`.

Tools are allowed to store state within their namespace of a `Report`.
If the tool has not yet been initialized, its `reset` function will be called.
You can update and access fields via dictionary access.

.. code-block:: python

    report[TOOL_NAME]['my flag'] = False

The `__init__.py` file for a Tool should use  `__all__` to expose any interesting
teacher-level functions, classes, and data. That way, teachers can just consistently use

.. code-block:: python

    from pedal.tool import *

To instantly gain access to the interesting set of members. The big exception to this is
:py:mod:`pedal.toolkit`, which is full of prototypical Tools organized into subsubmodules.
If it's just a single function without much complexity, it probably belongs in the Toolkit;
but as soon as complexity or multiple functions emerges, it can be promoted to a Tool.

"""

__all__ = ['Tool']


class Tool:
    pass



import sys

from pedal.resolvers import simple
from pedal.core.report import MAIN_REPORT


def resolve(report=None, priority_key=None):
    """
    Args:
        priority_key:
        report (Report): The report object to resolve down. Defaults to the
                         global MAIN_REPORT

    Returns
        str: A string of HTML feedback to be delivered
    """
    if report is None:
        report = MAIN_REPORT
    if priority_key is None:
        priority_key = simple.by_priority
    # Prepare feedbacks
    feedbacks = report.feedback
    feedbacks.sort(key=lambda f: (f.group or 0, priority_key(f)))
    suppressions = report.suppressions
    # Process
    final_success = False
    final_score = 0
    finals = {}
    found_failure = False
    for feedback in feedbacks:
        group = feedback.group or 0
        category = feedback.category.lower()
        if category in suppressions:
            if True in suppressions[category]:
                continue
            elif feedback.label.lower() in suppressions[category]:
                continue
        success, partial, message, data = simple.parse_feedback(feedback)
        final_success = success or final_success
        final_score += partial
        if message is not None:
            #print("RESETING GROUP", group, message[:20], found_failure, feedback.priority)
            if group not in finals:
                finals[group] = []
                found_failure = False
            if feedback.priority not in ('positive', 'instructions'):
                if found_failure:
                    continue
                found_failure = True
            entry = {'label': feedback.label,
                     'message': message,
                     'category': feedback.category,
                     'priority': feedback.priority,
                     'data': data}
            if feedback.priority == 'instructions':
                # Find end of instructions
                index = 0
                for feedback in finals[group]:
                    if feedback['priority'] != 'instructions':
                        break
                    index += 1
                finals[group].insert(index, entry)
            elif feedback.priority != 'positive':
                finals[group].insert(0, entry)
            else:
                finals[group].append(entry)
    #from pprint import pprint
    #pprint(finals)
    final_hide_correctness = suppressions.get('success', False)
    if not finals:
        finals[0] = [{
            'label': 'No errors',
            'category': 'Instructor',
            'data': [],
            'priority': 'medium',
            'message': "No errors reported."
        }]
    return (final_success, final_score, final_hide_correctness, finals)



from pedal.utilities.operators import OPERATION_DESCRIPTION

from pedal.core.commands import feedback
from pedal.core.feedback import AtomicFeedbackFunction
from pedal.core.report import MAIN_REPORT


def _generic_tifa_feedback_function(feedback_function, fields: dict, muted=False, report=MAIN_REPORT):
    return feedback(feedback_function.__name__, category=feedback.CATEGORIES.ALGORITHMIC,
                    message=feedback_function.message_template.format(**fields),
                    text=feedback_function.text_template.format(**fields),
                    fields=fields, kind=feedback.KINDS.MISTAKE, justification=feedback_function.justification,
                    tags=feedback_function.tags,
                    title=feedback_function.title, locations=fields['location'],
                    muted=muted or feedback_function.muted,
                    version=feedback_function.version, report=report)


@AtomicFeedbackFunction(title="Action after Return",
                        text_template=("You performed an action after already returning from a "
                                       "function, on line {location.line}. You can only return on a path "
                                       "once."),
                        justification=("TIFA visited a node not in the top scope when its "
                                       "*return variable was definitely set in this scope."))
def action_after_return(location, report=MAIN_REPORT):
    """

    Args:
        location:
        report:

    Returns:

    """
    fields = {'location': location}
    return _generic_tifa_feedback_function(action_after_return, fields, report)


@AtomicFeedbackFunction(title="Return outside Function",
                        text_template=("You attempted to return outside of a function on line {location.line}."
                                       " But you can only return from within a function."),
                        justification="TIFA visited a return node at the top level.")
def return_outside_function(location, report=MAIN_REPORT):
    """

    Args:
        location:
        report:

    Returns:

    """
    fields = {'location': location}
    return _generic_tifa_feedback_function(return_outside_function, fields, report)


@AtomicFeedbackFunction(title="Multiple Return Types",
                        text_template=(
                                "Your function returned {actual} on line {location.line}, even though you defined it to"
                                " return {expected}. Your function should return values consistently."
                        ),
                        justification="TIFA visited a function definition with multiple returns that unequal types.")
def multiple_return_types(location, expected, actual, report=MAIN_REPORT):
    """

    Args:
        location:
        expected:
        actual:
        report:

    Returns:

    """
    fields = {'location': location, 'expected': expected, 'actual': actual}
    return _generic_tifa_feedback_function(multiple_return_types, fields, report)


@AtomicFeedbackFunction(title="Write Out of Scope",
                        message_template=("You attempted to write the variable `{name}` from a higher scope "
                                          "(outside the function) on line {location.line}. You should only "
                                          "use variables inside the function they were declared in."
                                          ),
                        justification="TIFA stored to an existing variable not in this scope")
def write_out_of_scope(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(write_out_of_scope, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Unconnected Blocks",
                        text_template=("It looks like you have unconnected blocks on line {location.line}. "
                                       "Before you run your program, you must make sure that all "
                                       "of your blocks are connected that there are no unfilled "
                                       "holes."),
                        justification="TIFA found a name equal to ___")
def unconnected_blocks(location, report=MAIN_REPORT):
    """

    Args:
        location:
        report:

    Returns:

    """
    fields = {'location': location}
    return _generic_tifa_feedback_function(unconnected_blocks, fields, report=report)


@AtomicFeedbackFunction(title="Iteration Problem",
                        message_template=("The variable `{name}` was iterated on line "
                                          "{location.line} but you used the same variable as the iteration "
                                          "variable. You should choose a different variable name "
                                          "for the iteration variable. Usually, the iteration variable "
                                          "is the singular form of the iteration list (e.g., "
                                          "`for a_dog in dogs:`)."),
                        justification="TIFA visited a loop where the iteration list and target were the same.")
def iteration_problem(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(iteration_problem, fields, report=report)


@AtomicFeedbackFunction(title="Initialization Problem",
                        message_template=("The variable `{name}` was used on line {location.line}, "
                                          "but it was not given a value on a previous line. "
                                          "You cannot use a variable until it has been given a value."
                                          ),
                        justification="TIFA read a variable that did not exist or was not previously set in this branch.")
def initialization_problem(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(initialization_problem, fields, report=report)


@AtomicFeedbackFunction(title="Possible Initialization Problem",
                        message_template=("The variable `{name}` was used on line {location.line}, "
                                          "but it was possibly not given a value on a previous "
                                          "line. You cannot use a variable until it has been given "
                                          "a value. Check to make sure that this variable was "
                                          "declared in all of the branches of your decision."
                                          ),
                        justification="TIFA read a variable that was maybe set but not definitely set in this branch.")
def possible_initialization_problem(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(possible_initialization_problem, fields, report=report)


@AtomicFeedbackFunction(title="Unused Variable",
                        message_template=("The {kind} `{name}` was given a {initialization} on line {location.line}, but "
                                          "was never used after that."
                                          ),
                        justification="TIFA stored a variable but it was not read any other time in the program.")
def unused_variable(location, name, type, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        type:
        report:

    Returns:

    """
    if type.is_equal('function'):
        kind, initialization = 'function', 'definition'
    else:
        kind, initialization = 'variable', 'value'
    fields = {'location': location, 'name': name, 'type': type,
              'kind': kind, 'initialization': initialization}
    return _generic_tifa_feedback_function(unused_variable, fields, report=report)


@AtomicFeedbackFunction(title="Overwritten Variable",
                        message_template=("The variable `{name}` was given a value, but "
                                          "`{name}` was changed on line {location.line} before it "
                                          "was used. One of the times that you gave `{name}` "
                                          "a value was incorrect."
                                          ),
                        justification="TIFA attempted to store to a variable that was previously stored but not read.")
def overwritten_variable(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(overwritten_variable, fields, report=report)


@AtomicFeedbackFunction(title="Iterating over Non-list",
                        message_template=("The {iter} is not a list, but you used "
                                          "it in the iteration on line {location.line}. You should only iterate "
                                          "over sequences like lists."
                                          ),
                        justification="TIFA visited a loop's iteration list whose type was not indexable.")
def iterating_over_non_list(location, iter_name, report=MAIN_REPORT):
    """

    Args:
        location:
        iter_name:
        report:

    Returns:

    """
    if iter_name is None:
        iter_list = "expression"
    else:
        iter_list = "variable `{}`".format(iter_name)
    fields = {'location': location, 'name': iter_name, 'iter': iter_list}
    return _generic_tifa_feedback_function(iterating_over_non_list, fields, report=report)


@AtomicFeedbackFunction(title="Iterating over empty list",
                        message_template=("The {iter} was set as an empty list, "
                                          "and then you attempted to use it in an iteration on line "
                                          "{location.line}. You should only iterate over non-empty lists."
                                          ),
                        justification="TIFA visited a loop's iteration list that was empty.")
def iterating_over_empty_list(location, iter_name, report=MAIN_REPORT):
    """

    Args:
        location:
        iter_name:
        report:

    Returns:

    """
    if iter_name is None:
        iter_list = "expression"
    else:
        iter_list = "variable `{}`".format(iter_name)
    fields = {'location': location, 'name': iter_name, 'iter': iter_list}
    return _generic_tifa_feedback_function(iterating_over_empty_list, fields, report=report)


@AtomicFeedbackFunction(title="Incompatible types",
                        text_template=("You used {op_name} operation with {left_name} and {right_name} on line "
                                       "{location.line}. But you can't do that with that operator. Make "
                                       "sure both sides of the operator are the right type."
                                       ),
                        justification="TIFA visited an operation with operands of the wrong type.")
def incompatible_types(location, operation, left, right, report=MAIN_REPORT):
    """

    Args:
        location:
        operation:
        left:
        right:
        report:

    Returns:

    """
    op_name = OPERATION_DESCRIPTION.get(operation.__class__,
                                        str(operation))
    left_name = left.singular_name
    right_name = right.singular_name
    fields = {'location': location,
              'operation': operation, 'left': left, 'right': right,
              'op_name': op_name, 'left_name': left_name, 'right_name': right_name}
    return _generic_tifa_feedback_function(incompatible_types, fields, report=report)


@AtomicFeedbackFunction(title="Parameter Type Mismatch",
                        message_template=("You defined the parameter `{parameter_name}` on line {location.line} "
                                          "as {parameter_type_name}. However, the argument passed to that parameter "
                                          "was {argument_type_name}. The formal parameter type must match the argument's type."
                                          ),
                        justification="TIFA visited a function definition where a parameter type and argument type were not equal.")
def parameter_type_mismatch(location, parameter_name, parameter, argument, report=MAIN_REPORT):
    """

    Args:
        location:
        parameter_name:
        parameter:
        argument:
        report:

    Returns:

    """
    parameter_type_name = parameter.singular_name
    argument_type_name = argument.singular_name
    fields = {'location': location,
              'parameter_name': parameter_name,
              'parameter_type': parameter,
              'argument_type': argument,
              'parameter_type_name': parameter_type_name,
              'argument_type_name': argument_type_name}
    return _generic_tifa_feedback_function(parameter_type_mismatch, fields, report=report)


@AtomicFeedbackFunction(title="Read out of Scope",
                        message_template=("You attempted to read the variable `{name}` from a different scope on "
                                          "line {location.line}. You should only use variables inside the "
                                          "function they were declared in."
                                          ),
                        justification="TIFA read a variable that did not exist in this scope but existed in another.")
def read_out_of_scope(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(read_out_of_scope, fields, report=report)


# TODO: Complete these
@AtomicFeedbackFunction(title="Type Changes",
                        message_template="",
                        justification="",
                        muted=True)
def type_changes(location, name, old, new, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        old:
        new:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name, 'old': old, 'new': new}
    return _generic_tifa_feedback_function(type_changes, fields, report=report)


@AtomicFeedbackFunction(title="Unnecessary Second Branch",
                        message_template="",
                        justification="")
def unnecessary_second_branch(location, report=MAIN_REPORT):
    """

    Args:
        location:
        report:

    Returns:

    """
    fields = {'location': location}
    return _generic_tifa_feedback_function(unnecessary_second_branch, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Else on Loop Body",
                        message_template="",
                        justification="")
def else_on_loop_body(location, report=MAIN_REPORT):
    """

    Args:
        location:
        report:

    Returns:

    """
    fields = {'location': location}
    return _generic_tifa_feedback_function(else_on_loop_body, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Recursive Call",
                        message_template="",
                        justification="")
def recursive_call(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(recursive_call, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Not a Function",
                        message_template="",
                        justification="")
def not_a_function(location, name, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        report:

    Returns:

    """
    fields = {'location': location, 'name': name}
    return _generic_tifa_feedback_function(not_a_function, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Incorrect Arity",
                        message_template="",
                        justification="")
def incorrect_arity(location, function_name, report=MAIN_REPORT):
    """

    Args:
        location:
        function_name:
        report:

    Returns:

    """
    fields = {'location': location, 'function_name': function_name}
    return _generic_tifa_feedback_function(incorrect_arity, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Multiple Return Types",
                        message_template="",
                        justification="")
def multiple_return_types(location, expected, actual, report=MAIN_REPORT):
    """

    Args:
        location:
        expected:
        actual:
        report:

    Returns:

    """
    fields = {"location": location, "expected": expected, "actual": actual}
    return _generic_tifa_feedback_function(multiple_return_types, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Module Not Found",
                        text_template="",
                        justification="")
def module_not_found(location, name, is_dynamic=False, error=None, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        is_dynamic:
        error:
        report:

    Returns:

    """
    fields = {"location": location, "name": name, "is_dynamic": is_dynamic, "error": error}
    return _generic_tifa_feedback_function(module_not_found, fields, report=report, muted=True)


@AtomicFeedbackFunction(title="Append to non-list",
                        text_template="",
                        justification="",
                        muted=True)
def append_to_non_list(location, name, type, report=MAIN_REPORT):
    """

    Args:
        location:
        name:
        type:
        report:

    Returns:

    """
    fields = {'location': location, "name": name, "type": type }
    return _generic_tifa_feedback_function(append_to_non_list, fields, report=report)

'''
TODO: Finish these checks
"Empty Body": [], # Any use of pass on its own
"Malformed Conditional": [], # An if/else with empty else or if
"Unnecessary Pass": [], # Any use of pass
"Append to non-list": [], # Attempted to use the append method on a non-list
"Used iteration list": [], #
"Unused iteration variable": [], #
"Type changes": [], #
"Unknown functions": [], #
"Not a function": [], # Attempt to call non-function as function
"Recursive Call": [],
"Incorrect Arity": [],
"Aliased built-in": [], #
"Method not in Type": [], # A method was used that didn't exist for that type
"Submodule not found": [],
"Module not found": [],
"Else on loop body": [], # Used an Else on a For or While
'''

# TODO: Equality instead of assignment



# Skulpt has weird errors, and is missing some errors. Compatibility.
try:
    ParseError
except NameError:
    class ParseError(Exception):
        pass
try:
    SyntaxError
except NameError:
    class SyntaxError(Exception):
        pass
try:
    ReferenceError
except NameError:
    class ReferenceError(Exception):
        pass
try:
    EOFError
except NameError:
    class EOFError(Exception):
        pass
try:
    MemoryError
except NameError:
    class MemoryError(Exception):
        pass
try:
    OSError
except NameError:
    class OSError(Exception):
        pass
try:
    TokenError
except NameError:
    class TokenError(Exception):
        pass
try:
    TimeLimitError
except NameError:
    class TimeLimitError(Exception):
        pass

EXTENDED_ERROR_EXPLANATION = {
    ParseError: "A parse error means that Python does not understand the syntax on the line the error message points "
                "out. Common examples are forgetting commas beteween arguments or forgetting a <code>:</code> (colon) "
                "on a for statement.<br><b>Suggestion:</b> To fix a parse error you just need to look carefully at "
                "the line with the error and possibly the line before it.  Make sure it conforms to all of Python's "
                "rules.",
    TypeError: "Type errors most often occur when an expression tries to combine two objects with types that should "
               "not be combined. Like using <code>+</code> to add a number to a list instead of <code>.append</code>, "
               "or dividing a string by a number.<br><b>Suggestion:</b> To fix a type error you will most likely need "
               "to trace through your code and make sure the variables have the types you expect them to have.",
    SyntaxError: "This message indicates that Python can't figure out the syntax of a particular statement.  Some "
                 "examples are assigning to a literal, or a function call.      <br><b>Suggestion: </b>Check your "
                 "assignment statements and make sure that the left hand side of the assignment is a variable, "
                 "not a literal (e.g., 7 or \"hello\") or a function.",
    NameError: "A name error almost always means that you have used a variable before it has a value.  Often this may "
               "be a simple typo, so check the spelling carefully.  <br><b>Suggestion: </b>Check the right hand side "
               "of assignment statements and your function calls, this is the most likely place for a NameError to be "
               "found. It really helps to step through your code, one line at a time, mentally keeping track of your "
               "variables.",
    ValueError: "A ValueError most often occurs when you pass a parameter to a built-in function, and the function is "
                "expecting one type and you pass something  different. For instance, if you try to convert a "
                "non-numeric string to an int, you will get a ValueError:<br><pre>  int(\"Corgi\") # ValueError: "
                "invalid literal for int() with base 10</pre> <br><b>Suggestion: </b>The error message gives you a "
                "pretty good hint about the name of the function as well as the value that is incorrect.  Look at the "
                "error message closely and then trace back to the variable containing the problematic value. }",
    AttributeError: "This happens when you try to do <code>SOMETHING.WHATEVER</code> and either SOMETHING wasn't "
                    "declared or WHATEVER isn't an attribute of SOMETHING. This error message is telling you that the "
                    "object on the left hand side of the dot, does not have the attribute or method on the right hand "
                    "side.      <br><b>Suggestion: </b>You were probably trying to either get access to some data ("
                    "weather.get) or append (a_list.append). If it's the first one, you should make sure the module "
                    "is imported and that you are called its function correctly. If it's the second one, you should "
                    "make sure you spelled \"append\" right and that you are using a variable with a list for a "
                    "value.",
    TokenError: "Most of the time this error indicates that you have forgotten a right parenthesis or have forgotten "
                "to close a pair of quotes.  <br><b>Suggestion: </b>Check each line of your program and make sure "
                "that your parenthesis are balanced.",
    IndexError: "This message means that you are trying to index past the end of a string or a list.  For example, "
                "if your list has 3 things in it and you try to access the item at position 5.  <br><b>Suggestion: "
                "</b>Remember that the first item in a list or string is at index position 0, quite often this "
                "message comes about because you are off by one.  Remember in a list of length 3 the last legal index "
                "is 2.<br><pre>favorite_colors = [\"red\", \"blue\", \"green\"]\nfavorite_colors[2] # prints green "
                "favorite_color[3] # raises an IndexError</pre>",
    ImportError: "This error message indicates that you are trying to import a module that does not exist, or is not "
                 "in the same directory as your python script.  <br><b>Suggestion: </b>One problem may simply be that "
                 "you have a typo - remember, you must not capitalize the module name. Another common problem is that "
                 "you have placed the module in a different directory. Finally, if you're using a dataset module, "
                 "then it might not be imported. Use the \"Import Datasets\" button below!",
    ReferenceError: "This is a really hard error to get, so I'm not entirely sure what you did.  <br><b>Suggestion: "
                    "</b>Bring this code to the instructor. ",
    ZeroDivisionError: "This tells you that you are trying to divide by 0. Typically this is because the value of the "
                       "variable in the denominator of a division expression has the value 0.  <br><b>Suggestion: "
                       "</b>Are you sure you are dividing by the right variable? Are you sure that that variable has "
                       "the value you expect - is it possible that you counted the number of elements in an empty "
                       "list, for instance?",
    IndentationError: "This error occurs when you have not indented your code properly.  This is most likely to "
                      "happen as part of an if, for, while or def statement.  <br><b>Suggestion: </b>Check your if, "
                      "def, for, and while statements to be sure the lines are properly indented beneath them ("
                      "seriously, this happens ALL the time).  Another source of this error comes from copying and "
                      "pasting code where you have accidentally left some bits of code lying around that don't belong "
                      "there anymore. Finally, a very sinister but unlikely possibility is that you have some tab "
                      "characters in your code, which look identical to four spaces. Never, ever use tabs, "
                      "and carefully check code from the internet to make sure it doesn't have tabs.",
    EOFError: "If you are using input() or raw_input() commands, then this error happens when they don't get the "
              "right ending.  <br><b>Suggestion: </b>It's hard to protect against users. However, if you're using "
              "input(), you might be able to use raw_input() instead to avoid this problem. ",
    IOError: "This is a very easy error to get. The most common reason is that you were trying to open a file and it "
             "wasn't in the right place.   <br><b>Suggestion: </b>Make sure that the file is in the right place - "
             "print out the file path, and then check that it's definitely on your computer at that location. If you "
             "need help doing file processing, you should probably check with an instructor.",
    KeyError: "A dictionary has a bunch of keys that you can use to get data. This error is caused by you trying to "
              "refer to a key that does not exist.  <br><b>Suggestion: </b>The most common reason you get this "
              "exception is that you have a typo in your dictionary access. Check your spelling. Also double check "
              "that the key definitely exists.",
    MemoryError: "Somehow, you have run out of memory. <br><b>Suggestion: </b>Make sure you are filtering your "
                 "dataset! Alternatively, bring your code to an instructor.",
    OSError: "It's hard to say what an OSError is without deep checking. Many things can cause it.  "
             "<br><b>Suggestion: </b>Bring your code to an instructor.      ",
    TimeLimitError: "A TimeLimit error means that BlockPy wasn't able to process your program fast enough. Typically, "
                    "this means that you're iterating through too many elements. "
}
